import copy
import sys
import time


class MultiProgress:
    def __init__(self):
        self._bars = []
        self._lock = threading.Lock()  # Protects only the orchestrator's list structure
        self._running = False
        self._thread = None

    def add_bar(self, bar_instance: Bar) -> Bar:
        """User retains full, uninhibited ownership of the Bar instance."""
        with self._lock:
            self._bars.append(bar_instance)
        return bar_instance

    def _render_loop(self):
        self._set_low_priority()

        while self._running:
            # 1. Capture a structural snapshot of the registry list
            with self._lock:
                current_bars = self._bars[:]

            if not current_bars:
                time.sleep(0.04)  # Target ~24 FPS (1/24 ≈ 0.041)
                continue

            try:
                screen_cols, _ = Color.default().terminal_size()
            except Exception:
                screen_cols = 80

            lines = []
            for b in current_bars:
                # 2. Shallow copy the Bar object instantly on the orchestrator's time.
                # Captures current primitive value states (integers/strings).
                bar_snapshot = copy.copy(b)

                # 3. Process telemetry and extract the immutable rendering tuple
                render_tuple = bar_snapshot.call_render(screen_cols)
                lines.append("".join(render_tuple))

            # 4. Flush the complete immutable frame to the console
            sys.stdout.write("\n".join(lines))
            sys.stdout.flush()

            time.sleep(0.04)
            sys.stdout.write(f"\033[{len(lines)}A")


def run(self):
    def drpped_key(bar) -> bool:
        return bar.dropped

    while True:
        with self.lock:
            snapbars = sorted((copy.copy(b) for b in self.bars), dropped_key)
            # NOTE drop _after_ redrawing
            self.bars = [b for b in self.bars if not b.dropped]
        self.render_snapshot(snapbars)
        ...


def add_bar(self, bar):
    with self.lock:
        self.bars = [*self.bars, bar]
        # WARNING Do not drop before redrawing
        # self.bars = [b for b in [*self.bars, bar] if not b.dropped]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .contexts.infos import RuleInfo
from .contexts.memento import MEMENTO_DEFAULT_COLOR, memento
from .input import Cursor, LineInfo
from .ztyle import Color, Style


class TatSuException(Exception):
    pass


class ParseException(TatSuException):
    pass


class OptionSucceeded(ParseException):
    pass


class ParseError(ParseException):
    def __str__(self):
        msg = str(self.args[0]) if self.args else ''
        if not msg:
            return ''
        lines = msg.split('\n', 1)
        first = lines[0]
        rest = lines[1] if len(lines) > 1 else ''

        c = Color.stderr()
        err = Style(bold=True, fg=1, color=c)
        bold = Style(bold=True, color=c)

        result = f"{err('error:')} {bold(first)}"
        if rest:
            result += f"\n{rest}"
        return result


class HeartDied(ParseError):
    pass


class GrammarError(ParseError):
    pass


class CodegenError(ParseError):
    pass


class FailedSemantics(ParseException):
    pass


class FailedParse(ParseException):
    def __init__(self, cursor: Cursor, stack: list[RuleInfo], msg: str):
        # NOTE:
        #  Pass all arguments to super() to avoid pickling problems
        #  https://stackoverflow.com/a/28335286/545637
        super().__init__(cursor, stack, msg)

        self.cursor = cursor.clone()
        self.info: LineInfo = cursor.lineinfo()
        self.stack = [ri.name for ri in stack]
        self.msg = msg

    @property
    def pos(self) -> int:
        return self.cursor.pos

    @property
    def message(self):
        return self.msg

    def render(self, color: Color = MEMENTO_DEFAULT_COLOR) -> str:
        text = self.cursor.textstr
        msg = self.message
        info = self.info
        stack = self.stack

        return memento(msg, text, info, stack, color=color)

    def __str__(self):
        return self.render()


class FailedToken(FailedParse):
    def __init__(self, cursor: Cursor, stack: list[RuleInfo], token: str):
        super().__init__(cursor, stack, token)
        self.token = token

    @property
    def message(self):
        return 'expecting {}'.format(repr(self.token).lstrip('u'))


class FailedPattern(FailedParse):
    pass


class FailedMeta(FailedParse):
    pass


class FailedRef(FailedParse):
    def __init__(self, cursor: Cursor, stack: list[RuleInfo], name: str):
        super().__init__(cursor, stack, name)
        self.name = name

    @property
    def message(self):
        return f"Could not resolve reference to rule '{self.name}'"


class FailedChoice(FailedParse):
    @property
    def message(self):
        return 'No viable option'


class FailedLookahead(FailedParse):
    @property
    def message(self):
        return 'failed lookahead'


class FailedLeftRecursion(FailedParse):
    @property
    def message(self):
        return 'infinite left recursion'


class FailedExpectingEndOfText(FailedParse):
    pass


class FailedExpectingEndOfLine(FailedParse):
    pass


class KeywordError(FailedParse):
    pass


class FailedUnlinkedRule(FailedParse):
    pass


FailedKeywordSemantics = KeywordError



#!/usr/bin/env python3
#
#  CAVEAT UTILITOR
#
#  This file was automatically generated by 竜TatSu v5.23.1b1
#
#     https://pypi.python.org/pypi/tatsu/
#
#  Any changes you make to it will be overwritten the next time
#  the file is generated.
#
# noqa # type: ignore # ruff: noqa
# copyright: ignore
# fmt: off

from __future__ import annotations
from typing import Any

from tatsu.config import ParserConfig
from tatsu.contexts import CanParse
from tatsu.input import Text
from tatsu.input.buffer import Buffer
from tatsu.input.textlines import TextLines
from tatsu.peg import *


class TatSuBootstrapParser(CanParse):
    def __init__(self, config: Any = None, **settings: Any):
        self.config = ParserConfig.new(config, **settings)
        self.model = GRAMMAR_MODEL

    def parse(
        self,
        text: str | Text,
        /,
        *,
        start: str | None = None,
        asmodel: bool = True,
        config: Any = None,
        **settings: Any,
    ) -> Any:
        # NOTE
        #   Copy the grammar so the configuration is unique to this parse
        #   and one parse doesn't leak settings to another.
        #   There may be also configurations that are unique to this parse.
        model = Grammar(
            name=self.model.name,
            rules=self.model.rules,
            directives=self.model.directives,
            keywords=self.model.keywords,
            config=self.config,
        )
        config = ParserConfig.new(config, **settings)
        return model.parse(text, asmodel=asmodel, config=config)


GRAMMAR_MODEL: Grammar = (
    Grammar(
      name='TatSuBootstrap',
      directives={
        'grammar': 'TatSu',
        'comments': '(?ms)[(][*]\\s*(.*?)\\s*[*][)]|/[*]\\s*(.*?)\\s*[*]/',
        'eol_comments': '(?ms)(?:#|//)(.*?)$',
        'parseinfo': True,
        'left_recursion': False
      },
      keywords=(),
      rules=(
        Rule(
          name='start',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Call('grammar')
        ),
        Rule(
          name='grammar',
          params=['Grammar'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='title', exp=Constant('TATSU')),
              Closure(
                Choice(
                  [
                    NamedList(name='directives', exp=Call('directive')),
                    NamedList(name='keywords', exp=Call('keyword'))
                  ]
                )
              ),
              NamedList(name='rules', exp=Call('rule')),
              Closure(
                Choice(
                  [
                    NamedList(name='rules', exp=Call('rule')),
                    NamedList(name='keywords', exp=Call('keyword'))
                  ]
                )
              ),
              EOF()
            ]
          )
        ),
        Rule(
          name='directive',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Token('@@'),
              NegativeLookahead(Token('keyword')),
              Cut(),
              Group(
                Choice(
                  [
                    Sequence(
                      [
                        Named(name='name', exp=Group(Choice([Token('comments'), Token('eol_comments')]))),
                        Cut(),
                        Token('::'),
                        Cut(),
                        Named(name='value', exp=Call('regex'))
                      ]
                    ),
                    Sequence(
                      [
                        Named(name='name', exp=Token('whitespace')),
                        Cut(),
                        Token('::'),
                        Cut(),
                        Named(
                          name='value',
                          exp=Group(
                            Choice([Call('regex'), Call('string'), Token('None'), Token('False'), Constant()])
                          )
                        )
                      ]
                    ),
                    Sequence(
                      [
                        Named(
                          name='name',
                          exp=Group(
                            Choice(
                              [
                                Token('nameguard'),
                                Token('ignorecase'),
                                Token('left_recursion'),
                                Token('parseinfo'),
                                Token('memoization')
                              ]
                            )
                          )
                        ),
                        Cut(),
                        Group(
                          Choice(
                            [
                              Sequence([Token('::'), Cut(), Named(name='value', exp=Call('boolean'))]),
                              Named(name='value', exp=Constant(True))
                            ]
                          )
                        )
                      ]
                    ),
                    Sequence(
                      [
                        Named(name='name', exp=Token('grammar')),
                        Cut(),
                        Token('::'),
                        Cut(),
                        Named(name='value', exp=Call('word'))
                      ]
                    ),
                    Sequence(
                      [
                        Named(name='name', exp=Token('namechars')),
                        Cut(),
                        Token('::'),
                        Cut(),
                        Named(name='value', exp=Call('string'))
                      ]
                    )
                  ]
                )
              ),
              Cut()
            ]
          )
        ),
        Rule(
          name='keywords',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=PositiveClosure(Call('keyword'))
        ),
        Rule(
          name='keyword',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Token('@@keyword'),
              Cut(),
              Token('::'),
              Group(
                Choice(
                  [
                    Sequence(
                      [
                        Token('('),
                        Cut(),
                        Override(PositiveClosure(Choice([Call('word'), Call('string')]))),
                        Token(')')
                      ]
                    ),
                    PositiveClosure(
                      Sequence(
                        [
                          OverrideList(Group(Choice([Call('word'), Call('string')]))),
                          NegativeLookahead(Group(Choice([Token(':'), Token('=')])))
                        ]
                      )
                    )
                  ]
                )
              )
            ]
          )
        ),
        Rule(
          name='params',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              OverrideList(Call('first_param')),
              Closure(
                Sequence(
                  [
                    Token(','),
                    OverrideList(Call('literal')),
                    NegativeLookahead(Token('=')),
                    Cut()
                  ]
                )
              )
            ]
          )
        ),
        Rule(
          name='first_param',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice([Call('path'), Call('literal')])
        ),
        Rule(
          name='kwparams',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=PositiveGather(exp=Call('pair'), sep=Token(','))
        ),
        Rule(
          name='the_params_at_last',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Named(name='kwparams', exp=Call('kwparams')),
              Sequence(
                [
                  Named(name='params', exp=Call('params')),
                  Token(','),
                  Cut(),
                  Named(name='kwparams', exp=Call('kwparams'))
                ]
              ),
              Named(name='params', exp=Call('params'))
            ]
          )
        ),
        Rule(
          name='paramdef',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Sequence([Token('['), Cut(), RuleInclude('the_params_at_last'), Token(']')]),
              Sequence([Token('('), Cut(), RuleInclude('the_params_at_last'), Token(')')]),
              Sequence([Token('::'), Cut(), Named(name='params', exp=Call('params'))])
            ]
          )
        ),
        Rule(
          name='rule',
          params=['Rule'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='decorators', exp=Closure(Call('decorator'))),
              Named(name='name', exp=NameMeta()),
              Cut(),
              Optional(RuleInclude('paramdef')),
              Optional(Sequence([Token('<'), Cut(), Named(name='base', exp=Call('known_name'))])),
              Void(),
              Pattern('=|::=|:=?'),
              Cut(),
              Named(name='exp', exp=Call('expre')),
              Call('ENDRULE'),
              Cut()
            ]
          )
        ),
        Rule(
          name='ENDRULE',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=True,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice([Call('DEDENT'), Call('BLANK'), Token(';'), EOF()])
        ),
        Rule(
          name='DEDENT',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=True,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Call('EOL'), Lookahead(Pattern('^\\S'))])
        ),
        Rule(
          name='BLANK',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=True,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Call('EOL'), Call('EOL')])
        ),
        Rule(
          name='EOL',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=True,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Pattern('(?m)[ \\t]*$'), Pattern('(?m)(?:\\r?\\n|\\r)')])
        ),
        Rule(
          name='decorator',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Token('@'),
              NegativeLookahead(Token('@')),
              Cut(),
              Override(Pattern('(override|name|isname|nomemo|nostak)\\b'))
            ]
          )
        ),
        Rule(
          name='pair',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [OverrideList(Call('word')), Token('='), Cut(), OverrideList(Call('literal'))]
          )
        ),
        Rule(
          name='expre',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice([Call('choice'), Call('sequence')])
        ),
        Rule(
          name='choice',
          params=['Choice'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Optional(Sequence([Token('|'), Cut()])),
              OverrideList(Call('option')),
              PositiveClosure(Sequence([Token('|'), Cut(), OverrideList(Call('option'))]))
            ]
          )
        ),
        Rule(
          name='option',
          params=['Option'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Call('sequence')
        ),
        Rule(
          name='sequence',
          params=['Sequence'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Sequence(
                [
                  Lookahead(Group(Sequence([Call('element'), Token(',')]))),
                  PositiveGather(exp=Call('element'), sep=Token(','))
                ]
              ),
              PositiveClosure(Sequence([NegativeLookahead(Call('ENDRULE')), Call('element')]))
            ]
          )
        ),
        Rule(
          name='element',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Call('override'),
              Call('meta'),
              Call('named'),
              Call('term'),
              Call('rule_include')
            ]
          )
        ),
        Rule(
          name='rule_include',
          params=['RuleInclude'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Token('>'), Cut(), Override(Call('known_name'))])
        ),
        Rule(
          name='named',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice([Call('named_list'), Call('named_single')])
        ),
        Rule(
          name='named_list',
          params=['NamedList'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='name', exp=NameMeta()),
              Pattern('\\+[:=]'),
              Cut(),
              Named(name='exp', exp=Call('term'))
            ]
          )
        ),
        Rule(
          name='named_single',
          params=['Named'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='name', exp=NameMeta()),
              Pattern('[:=]'),
              Cut(),
              Named(name='exp', exp=Call('term'))
            ]
          )
        ),
        Rule(
          name='override',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice([Call('override_list'), Call('override_single')])
        ),
        Rule(
          name='override_list',
          params=['OverrideList'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Pattern('\\+=|@\\+:'), Cut(), Override(Call('term'))])
        ),
        Rule(
          name='override_single',
          params=['Override'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Pattern('=|@:'), Cut(), Override(Call('term'))])
        ),
        Rule(
          name='term',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Call('gather'),
              Call('join'),
              Call('left_join'),
              Call('right_join'),
              Call('empty_closure'),
              Call('positive_closure'),
              Call('closure'),
              Call('optional'),
              Call('atom'),
              Call('void'),
              Call('skip_to'),
              Call('lookahead'),
              Call('negative_lookahead'),
              Call('cut'),
              Call('cut_deprecated')
            ]
          )
        ),
        Rule(
          name='group',
          params=['Group'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              NegativeLookahead(Token('(?:')),
              Token('('),
              Cut(),
              Override(Call('expre')),
              Token(')'),
              Cut()
            ]
          )
        ),
        Rule(
          name='skip',
          params=['SkipGroup'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Token('(?:'), Cut(), Override(Call('expre')), Token(')'), Cut()])
        ),
        Rule(
          name='gather',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Lookahead(Group(Sequence([Call('atom'), Token('.{')]))),
              Cut(),
              Group(Choice([Call('positive_gather'), Call('normal_gather')]))
            ]
          )
        ),
        Rule(
          name='positive_gather',
          params=['PositiveGather'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='sep', exp=Call('atom')),
              Token('.{'),
              Named(name='exp', exp=Call('expre')),
              Token('}'),
              NegativeLookahead(Pattern('\\+=')),
              Pattern('[+-]'),
              Cut()
            ]
          )
        ),
        Rule(
          name='normal_gather',
          params=['Gather'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='sep', exp=Call('atom')),
              Token('.{'),
              Cut(),
              Named(name='exp', exp=Call('expre')),
              Token('}'),
              Optional(Sequence([Token('*'), Cut()])),
              Cut()
            ]
          )
        ),
        Rule(
          name='join',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Lookahead(Group(Sequence([Call('atom'), Token('%{')]))),
              Cut(),
              Group(Choice([Call('positive_join'), Call('normal_join')]))
            ]
          )
        ),
        Rule(
          name='positive_join',
          params=['PositiveJoin'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='sep', exp=Call('atom')),
              Token('%{'),
              Named(name='exp', exp=Call('expre')),
              Token('}'),
              NegativeLookahead(Pattern('\\+=')),
              Pattern('[+-]'),
              Cut()
            ]
          )
        ),
        Rule(
          name='normal_join',
          params=['Join'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='sep', exp=Call('atom')),
              Token('%{'),
              Cut(),
              Named(name='exp', exp=Call('expre')),
              Token('}'),
              Optional(Sequence([Token('*'), Cut()])),
              Cut()
            ]
          )
        ),
        Rule(
          name='left_join',
          params=['LeftJoin'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='sep', exp=Call('atom')),
              Token('<{'),
              Cut(),
              Named(name='exp', exp=Call('expre')),
              Token('}'),
              NegativeLookahead(Pattern('\\+=')),
              Pattern('[+-]'),
              Cut()
            ]
          )
        ),
        Rule(
          name='right_join',
          params=['RightJoin'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='sep', exp=Call('atom')),
              Token('>{'),
              Cut(),
              Named(name='exp', exp=Call('expre')),
              Token('}'),
              NegativeLookahead(Pattern('\\+=')),
              Pattern('[+-]'),
              Cut()
            ]
          )
        ),
        Rule(
          name='positive_closure',
          params=['PositiveClosure'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Sequence(
                [
                  Token('{'),
                  Override(Call('expre')),
                  Token('}'),
                  NegativeLookahead(Pattern('\\+=')),
                  Pattern('[+-]'),
                  Cut()
                ]
              ),
              Sequence(
                [
                  Override(Call('atom')),
                  NegativeLookahead(Pattern('\\+=')),
                  Pattern('[+]'),
                  Cut()
                ]
              )
            ]
          )
        ),
        Rule(
          name='closure',
          params=['Closure'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Sequence(
                [Token('{'), Override(Call('expre')), Token('}'), Optional(Token('*')), Cut()]
              ),
              Sequence([Override(Call('atom')), Token('*'), Cut()])
            ]
          )
        ),
        Rule(
          name='empty_closure',
          params=['EmptyClosure'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Token('{}'), Cut(), Override(Void())])
        ),
        Rule(
          name='optional',
          params=['Optional'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Sequence([Token('['), Cut(), Override(Call('expre')), Token(']'), Cut()]),
              Sequence(
                [
                  Override(Call('atom')),
                  NegativeLookahead(Group(Choice([Token('?"'), Token("?'"), Token('?/')]))),
                  Token('?'),
                  Cut()
                ]
              )
            ]
          )
        ),
        Rule(
          name='lookahead',
          params=['Lookahead'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Token('&'), Cut(), Override(Call('term'))])
        ),
        Rule(
          name='negative_lookahead',
          params=['NegativeLookahead'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Token('!'), Cut(), Override(Call('term'))])
        ),
        Rule(
          name='skip_to',
          params=['SkipTo'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Token('->'), Cut(), Override(Call('term'))])
        ),
        Rule(
          name='atom',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Call('meta'),
              Call('token'),
              Call('call'),
              Call('dot'),
              Call('pattern'),
              Call('skip'),
              Call('group'),
              Call('eol'),
              Call('eof'),
              Call('alert'),
              Call('constant')
            ]
          )
        ),
        Rule(
          name='meta',
          params=['Meta'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Pattern('@(name|int|uint|float|bool)\\b')
        ),
        Rule(
          name='call',
          params=['Call'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=NameMeta()
        ),
        Rule(
          name='void',
          params=['Void'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Token('()'), Cut()])
        ),
        Rule(
          name='fail',
          params=['Fail'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Token('!()'), Cut()])
        ),
        Rule(
          name='cut',
          params=['Cut'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Token('~'), Cut()])
        ),
        Rule(
          name='cut_deprecated',
          params=['Cut'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Token('>>'), Cut()])
        ),
        Rule(
          name='known_name',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([NameMeta(), Cut()])
        ),
        Rule(
          name='name',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Call('word')
        ),
        Rule(
          name='constant',
          params=['Constant'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Lookahead(Token('`')),
              Group(
                Choice(
                  [
                    Pattern('(?ms)```((?:.|\\n)*?)```'),
                    Sequence([Token('`'), Override(Call('literal')), Token('`')]),
                    Pattern('`(.*?)`')
                  ]
                )
              )
            ]
          )
        ),
        Rule(
          name='alert',
          params=['Alert'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='level', exp=Pattern('\\^+')),
              Named(name='message', exp=Call('constant'))
            ]
          )
        ),
        Rule(
          name='token',
          params=['Token'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice([Call('string'), Call('raw_string')])
        ),
        Rule(
          name='literal',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Call('raw_string'),
              Call('value'),
              Call('boolean'),
              Call('none'),
              Call('word'),
              Call('hex'),
              Call('float'),
              Call('int')
            ]
          )
        ),
        Rule(
          name='string',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Lookahead(Group(Choice([Token('"'), Token("'")]))),
              Group(
                Choice([Call('multiline_string'), Call('singlequoted'), Call('doublequoted')])
              )
            ]
          )
        ),
        Rule(
          name='singlequoted',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Call('SINGLEQUOTED')
        ),
        Rule(
          name='doublequoted',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Call('DOUBLEQUOTED')
        ),
        Rule(
          name='raw_string',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Lookahead(Pattern('r["\']')), Pattern('r'), Override(Call('STRING'))])
        ),
        Rule(
          name='STRING',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=True,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice([Call('SINGLEQUOTED'), Call('DOUBLEQUOTED')])
        ),
        Rule(
          name='SINGLEQUOTED',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=True,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Pattern("'((?:[^'\\n]|\\\\'|\\\\\\\\)*?)'"), Cut()])
        ),
        Rule(
          name='DOUBLEQUOTED',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=True,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Pattern('"((?:[^"\\n]|\\\\"|\\\\\\\\)*?)"'), Cut()])
        ),
        Rule(
          name='multiline_string',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Sequence([Pattern("(?ms)'''((?:\\\\\\\\|\\\\.|.)*?)'''"), Cut()]),
              Sequence([Pattern('(?ms)"""((?:\\\\\\\\|\\\\.|.)*?)"""'), Cut()])
            ]
          )
        ),
        Rule(
          name='hex',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Pattern('0[xX](?:\\d|[a-fA-F])+')
        ),
        Rule(
          name='float',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Pattern('[-+]?(?:\\d+\\.\\d*|\\d*\\.\\d+)(?:[Ee][-+]?\\d+)?')
        ),
        Rule(
          name='int',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Pattern('[-+]?\\d+')
        ),
        Rule(
          name='path',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Pattern('[_\\w][_\\w\\d]*(?:::[_\\w][_\\w\\d]*)+')
        ),
        Rule(
          name='word',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Pattern('(?ms)\\s*([_\\w][_\\w\\d]*)\\b')
        ),
        Rule(
          name='dot',
          params=['Dot'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Token('/./')
        ),
        Rule(
          name='pattern',
          params=['Pattern'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Call('regex')
        ),
        Rule(
          name='regex',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Call('deprecated_regex'),
              Sequence(
                [
                  NegativeLookahead(Token('?/')),
                  Group(Choice([Call('REGEX'), Sequence([Token('?'), Override(Call('STRING'))])]))
                ]
              )
            ]
          )
        ),
        Rule(
          name='REGEX',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=True,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [Lookahead(Token('/')), Pattern('(?ms)/((?:[^/\\\\]|\\\\/|\\\\.)*)/'), Cut()]
          )
        ),
        Rule(
          name='deprecated_regex',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence(
            [
              Token('?/'),
              Cut(),
              Override(Pattern('(?ms)((?:[^/\\\\]|\\\\/|\\\\.)*)')),
              Cut(),
              Token('/?')
            ]
          )
        ),
        Rule(
          name='boolean',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice([Token('True'), Token('False')])
        ),
        Rule(
          name='none',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Token('None')
        ),
        Rule(
          name='eof',
          params=['EOF'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Token('$'), Cut()])
        ),
        Rule(
          name='eol',
          params=['EOL'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Token('$->')
        ),
        Rule(
          name='value',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice([Call('string'), Call('number'), Call('true'), Call('false'), Call('null')])
        ),
        Rule(
          name='number',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Pattern('-?(?:0|[1-9][0-9]*)(?:\\.[0-9]+)?(?:[eE][+-]?[0-9]+)?')
        ),
        Rule(
          name='true',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Token('true')
        ),
        Rule(
          name='false',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Token('false')
        ),
        Rule(
          name='null',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Token('null')
        )
      )
    )
)


class TatSuBootstrapText(TextLines):
    def __init__(
        self,
        text,
        /,
        config: ParserConfig | None = None,
        **settings,
    ) -> None:
        config = ParserConfig.new(
            config=config,
            name='TatSu',
            whitespace=None,
            nameguard=None,
            ignorecase=False,
            namechars='',
            parseinfo=True,
            comments=r'(?ms)[(][*]\s*(.*?)\s*[*][)]|/[*]\s*(.*?)\s*[*]/',
            eol_comments=r'(?ms)(?:#|//)(.*?)$',
            keywords=KEYWORDS,
            start='start',
        )
        assert isinstance(config, ParserConfig)
        config = config.override(**settings)

        super().__init__(text, config=config)

TatSuBootstrapTokenizer = TatSuBootstrapText


class TatSuBootstrapBuffer(Buffer):  # NOTE: backwards compatibility
    def __init__(
        self,
        text,
        /,
        config: ParserConfig | None = None,
        **settings,
    ) -> None:
        config = ParserConfig.new(
            config=config,
            name='TatSu',
            whitespace=None,
            nameguard=None,
            ignorecase=False,
            namechars='',
            parseinfo=True,
            comments=r'(?ms)[(][*]\s*(.*?)\s*[*][)]|/[*]\s*(.*?)\s*[*]/',
            eol_comments=r'(?ms)(?:#|//)(.*?)$',
            keywords=KEYWORDS,
            start='start',
        )
        assert isinstance(config, ParserConfig)
        config = config.override(**settings)

        super().__init__(text, config=config)


KEYWORDS = ()


def main(filename, **kwargs):
    if not filename or filename == '-':
        import sys

        text = sys.stdin.read()
    else:
        import pathlib

        text = pathlib.Path(filename).read_text()

    parser = TatSuBootstrapParser()
    return parser.parse(text, filename=filename, **kwargs)


if __name__ == '__main__':
    import json
    from tatsu.parsing import generic_main
    from tatsu.util import asjson

    ast = generic_main(main, TatSuBootstrapParser, name='TatSuBootstrap')
    data = asjson(ast)
    print(json.dumps(data, indent=2))



import asyncio
import concurrent.futures
from pathlib import Path


def heavy_computation(x: int):
    # This runs in a separate process, bypassing the GIL
    return sum(i * i for i in range(10**7)) + x


async def main():
    loop = asyncio.get_running_loop()

    # We use a ProcessPoolExecutor for true parallel execution
    with concurrent.futures.ProcessPoolExecutor() as pool:
        # Create a list of tasks
        tasks = [loop.run_in_executor(pool, heavy_computation, i) for i in range(4)]

        # Await all parallel processes
        results = await asyncio.gather(*tasks)
        print(f"Results: {results}")


if __name__ == "__main__":
    asyncio.run(main())



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from tatsu.api import compile
from tatsu.util.asjson import asjsons  # noqa: F401


def test_alert_interpolation(trace=False):
    input = '42 69'
    grammar = r"""
            start = a:number b: number i:^`"seen: {a}, {b}"` $ ;
            number::int = /\d+/ ;
    """
    model = compile(grammar)
    # print(asjsons(model))
    ast = model.parse(input, trace=trace)
    assert ast == {'a': '42', 'b': '69', 'i': None}
    ast = model.parse(input, asmodel=True, trace=trace)
    assert ast == {'a': 42, 'b': 69, 'i': None}



from __future__ import annotations

from unittest.mock import patch

import pytest

from tatsu.ztyle.markup import markup, tokenize, tokenize_to_values


@pytest.fixture(autouse=True)  # noqa
def _tty_stdout():
    with patch("sys.stdout.isatty", return_value=True):
        yield


def test_plain_text():
    r = markup("hello")
    assert len(r.style) == 1
    assert r.style[0].value == "hello"


def test_plain_text_with_spaces():
    r = markup("hello world")
    assert len(r.style) == 1
    assert r.style[0].value == "hello world"


def test_single_bold():
    r = markup("[bold]hello[/bold]")
    assert len(r.style) == 1
    assert r.style[0].value == "hello"
    assert r.style[0]._bold is True


def test_single_italic():
    r = markup("[italic]hello[/italic]")
    assert len(r.style) == 1
    assert r.style[0].value == "hello"
    assert r.style[0]._italic is True


def test_single_color():
    r = markup("[red]hello[/red]")
    assert len(r.style) == 1
    assert r.style[0].value == "hello"
    assert r.style[0]._fg == 1


def test_multiple_attributes_one_tag():
    r = markup("[bold red]hello[/]")
    assert len(r.style) == 1
    assert r.style[0].value == "hello"
    assert r.style[0]._bold is True
    assert r.style[0]._fg == 1


def test_mixed_plain_and_styled():
    r = markup("plain [bold]bold[/bold] plain")
    assert len(r.style) == 3
    assert r.style[0].value == "plain "
    assert r.style[1].value == "bold"
    assert r.style[1]._bold is True
    assert r.style[2].value == " plain"


def test_styled_then_plain():
    r = markup("[bold]bold[/bold] plain")
    assert len(r.style) == 2
    assert r.style[0].value == "bold"
    assert r.style[0]._bold is True
    assert r.style[1].value == " plain"


def test_plain_then_styled():
    r = markup("plain [bold]bold[/]")
    assert len(r.style) == 2
    assert r.style[0].value == "plain "
    assert r.style[1].value == "bold"
    assert r.style[1]._bold is True


def test_nested_styles():
    r = markup("[bold]bold [red]red[/red] only bold[/bold]")
    assert len(r.style) == 3
    assert r.style[0].value == "bold "
    assert r.style[0]._bold is True
    assert r.style[0]._fg == -1
    assert r.style[1].value == "red"
    assert r.style[1]._bold is True
    assert r.style[1]._fg == 1
    assert r.style[2].value == " only bold"
    assert r.style[2]._bold is True
    assert r.style[2]._fg == -1


def test_close_all():
    r = markup("[bold red]hello[/all]world")
    assert len(r.style) == 2
    assert r.style[0].value == "hello"
    assert r.style[0]._bold is True
    assert r.style[0]._fg == 1
    assert r.style[1].value == "world"
    assert r.style[1]._bold is False
    assert r.style[1]._fg == -1


def test_escape_brackets():
    r = markup("[[hello]]")
    assert len(r.style) == 1
    assert r.style[0].value == "[hello]]"


def test_empty_string():
    r = markup("")
    assert len(r.style) == 0
    assert r.style == []


def test_color_methods():
    r = markup("[pink]hello[/]")
    assert len(r.style) == 1
    assert r.style[0].value == "hello"
    assert r.style[0]._fg == 200  # pink is ANSI 256-color 200


def test_consecutive_styles():
    r = markup("[bold]a[/][italic]b[/]")
    assert len(r.style) == 2
    assert r.style[0].value == "a"
    assert r.style[0]._bold is True
    assert r.style[0]._italic is False
    assert r.style[1].value == "b"
    assert r.style[1]._bold is False
    assert r.style[1]._italic is True


def test_tokenize_plain():
    toks = tokenize("hello")
    assert toks[0].value == "hello"


def test_tokenize_tag():
    toks = tokenize_to_values("[bold]hello[/]")
    assert toks == ["bold", "hello", ""]


def test_tokenize_escaped():
    toks = tokenize_to_values("[[hello]]")
    assert toks == ["[", "hello]]"]


def test_tokenize_multiple_attrs():
    toks = tokenize("[bold red]x[/]")
    assert toks[0].value == "bold red"



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
# by Gemini 2026-03-07
from __future__ import annotations

from typing import Any

import pytest

from tatsu.util import cast


def test_cast_success():
    assert cast(int, 42) == 42
    assert cast(str, 'hello') == 'hello'

    assert cast(list[int], [1, 2, 3]) == [1, 2, 3]
    assert cast(dict[str, int], {'a': 1}) == {'a': 1}


@pytest.mark.parametrize(
    'target, value, expected_name',
    [
        (int, 'not an int', 'int'),
        (list[int], {'not': 'a list'}, 'list[int]'),
        (float, None, 'float'),
        (dict[str, Any], ['a'], 'dict[str, Any]'),
    ],
)
def test_cast_failure(target: type, value: Any, expected_name: str):
    with pytest.raises(TypeError) as info:
        cast(target, value)

    # Verify the recovered type name is in the error message
    assert expected_name in str(info.value)
    assert type(value).__name__ in str(info.value)


def test_cast_pathlib():
    from pathlib import Path

    path = Path('test.txt')
    result = cast(Path, path)
    assert isinstance(result, Path)


@pytest.mark.parametrize(
    "target, value",
    [
        (int | str, 42),  # Success with first type
        (int | str, "hello"),  # Success with second type
        (float | None, None),  # Success with NoneType
        (list | dict, [1, 2]),  # Success with collection union
    ],
)
def test_cast_union_success(target: type, value: Any):
    # Should not raise any exception
    assert cast(target, value) == value


@pytest.mark.parametrize(
    "target, value, expected_msg",
    [
        (int | str, 1.5, "Expected int | str, but got float instead"),
        (bool | int, "True", "Expected bool | int, but got str instead"),
        (list[int] | str, 42, "Expected list[int] | str, but got int instead"),
    ],
)
def test_cast_union_failure(target: type, value: Any, expected_msg: str):
    with pytest.raises(TypeError) as info:
        cast(target, value)

    # Verify the regex-cleaned 'expected' string matches our message
    assert expected_msg in str(info.value)



#  Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
#  SPDX-License-Identifier: BSD-4-Clause
#

# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import re
from typing import Any

from .. import peg as g
from ..contexts import AST


def camel2py(name: Any) -> str:
    return re.sub(
        r'([a-z0-9])([A-Z])',
        lambda m: m.group(1) + '_' + m.group(2).lower(),
        str(name),
    )


class ANTLRSemantics:
    def __init__(self, name):
        self.name = name
        self.tokens = {}
        self.token_rules = {}
        self.synthetic_rules = []

    def grammar(self, ast: AST) -> g.Grammar:
        # Add any defined tokens that were not part of a rule
        for name, exp in self.tokens.items():
            if name not in self.token_rules:
                self.synthetic_rules.append(g.Rule(name=name, exp=exp))

        # For non-trivial token rules (non-Token expressions), use Fail()
        # since converting ANTLR patterns to Python regex is not viable.
        for rule in ast.rules:
            if rule is None:
                continue
            token_name = rule.name.upper() if rule.name else ''
            if token_name in self.token_rules and not isinstance(
                self.token_rules.get(token_name), g.Synth
            ):
                exp = self.token_rules[token_name]
                if not isinstance(exp, g.Token):
                    rule.exp = g.Fail()

        # Resolve or fail any remaining placeholders
        for name, exp in self.token_rules.items():
            if isinstance(exp, g.Synth):
                # This placeholder was referenced but never defined.
                # Create a new rule that fails.
                rule_name = name
                if not any(r.name == rule_name for r in ast.rules):
                    self.synthetic_rules.append(g.Rule(name=rule_name, exp=g.Fail()))

        # Deduplicate rules by name, keeping the first occurrence
        all_rules = [r for r in (*ast.rules, *self.synthetic_rules) if r is not None]
        seen: set[str] = set()
        deduped = []
        for r in all_rules:
            if r.name not in seen:
                seen.add(r.name)
                deduped.append(r)
        return g.Grammar(self.name, deduped)

    def rule(self, ast: AST) -> g.Rule | None:
        name = camel2py(ast.name)
        exp = ast.exp
        if name[0].isupper():
            name = name.upper()
            old = self.token_rules.get(name)
            self.token_rules[name] = exp
            if isinstance(exp, g.Token) and isinstance(old, g.Synth):
                old.exp = exp
            return g.Rule(name=name, exp=exp)

        return g.Rule(name=name, exp=exp)

    def alternatives(self, ast: AST) -> g.Model:
        options = [o for o in ast.options if o is not None]
        if len(options) == 1:
            return options[0]
        else:
            options = [
                g.Option(
                    exp=g.Group(exp=o) if isinstance(o, (g.Choice, g.Sequence)) else o
                )
                for o in options
            ]
            return g.Choice(options=options)

    def elements(self, ast: list[g.Model]) -> g.Model:
        elements = [e for e in ast if e is not None]
        if not elements:
            return g.Void()
        elif len(elements) == 1:
            return elements[0]
        else:
            return g.Sequence(sequence=elements)

    def predicate_or_action(self, _ast: Any) -> None:
        return None

    def named(self, ast: AST) -> g.Named:
        if ast.force_list:
            return g.NamedList(name=ast.name, exp=ast.exp)
        else:
            return g.Named(name=ast.name, exp=ast.exp)

    def syntactic_predicate(self, _ast: Any) -> None:
        return None

    def optional(self, ast: g.Model) -> g.Optional:
        if isinstance(ast, g.Group):
            ast = ast.exp
        return g.Optional(exp=ast)

    def closure(self, ast: g.Model) -> g.Closure:
        if isinstance(ast, g.Group):
            ast = ast.exp
        return g.Closure(exp=ast)

    def positive_closure(self, ast: g.Model) -> g.Closure:
        if isinstance(ast, g.Group):
            ast = ast.exp
        return g.PositiveClosure(exp=ast)

    def negative(self, ast: g.Model) -> g.Sequence:
        neg = g.NegativeLookahead(exp=ast)
        any = g.Pattern(pattern='.')
        return g.Sequence(sequence=[neg, any])

    def subexp(self, ast: g.Model) -> g.Group:
        return g.Group(exp=ast)

    def regexp(self, ast: AST) -> g.Pattern:
        pattern = ''.join(ast)
        regex = re.compile(pattern)
        return g.Pattern(regex.pattern)

    def charset_optional(self, ast: AST) -> str:
        return f'{ast}?'

    def charset_closure(self, ast: AST) -> str:
        return f'{ast}*'

    def charset_positive_closure(self, ast: AST) -> str:
        return f'{ast}+'

    def charset_or(self, ast: AST) -> str:
        return '[{}]'.format(''.join(ast))

    def charset_negative_or(self, ast: AST) -> str:
        return '[^{}]'.format(''.join(ast))

    @staticmethod
    def escape(s: str) -> str:
        return ''.join('\\' + c if c in '[]().*+{}^$' else c for c in s)

    def charset_atom(self, ast: AST) -> AST:
        return ast

    def charset_char(self, ast: AST) -> AST:
        return ast

    def charset_range(self, ast: AST) -> str:
        return f'{ast.first}-{ast.last}'

    def newranges(self, ast: AST) -> g.Pattern:
        pattern = ''.join(ast)
        regex = re.compile(pattern)
        return g.Pattern(regex.pattern)

    def newrange(self, ast: AST) -> str:
        pattern = '[{}]{}'.format(ast.range, ast.repeat or '')
        re.compile(pattern)
        return pattern

    def negative_newrange(self, ast: AST) -> str:
        pattern = '[^{}]{}'.format(ast.range, ast.repeat or '')
        re.compile(pattern)
        return pattern

    def call(self, ast: str) -> g.Call:
        assert ast[0].islower()
        return g.Call(name=camel2py(ast))

    def any(self, _ast: AST) -> g.Pattern:
        return g.Pattern(r'\w+|\S+')

    def string(self, ast: AST) -> g.Token:
        text = str(ast)
        if isinstance(text, list):
            text = ''.join(text)
        return g.Token(token=text)

    def eof(self, _ast: AST) -> g.EOF:
        return g.EOF()

    def token(self, ast: AST) -> g.Token | g.Void:
        name = ast.name
        if ast.exp:
            exp = g.Token(token=ast.value)
            self.tokens[name] = exp
        else:
            exp = g.Void()  # type: ignore
            rule = g.Rule(exp=exp, name=name)
            self.synthetic_rules.append(rule)
        return exp

    def token_ref(self, ast: AST) -> g.Model:
        name = camel2py(ast).upper()

        if name in self.token_rules and not isinstance(self.token_rules[name], g.Synth):
            result = self.token_rules[name]
            if isinstance(result, g.Token):
                return result
            return g.Call(name=name)

        if name in self.token_rules:
            return self.token_rules[name]

        exp = g.Synth(exp=g.Call(name=name))
        self.token_rules[name] = exp
        return exp



#!/usr/bin/env python
"""
Solution to Project Euler Problem
http://projecteuler.net/

by Apalala <apalala@gmail.com>
(cc) Attribution-ShareAlike
http://creativecommons.org/licenses/by-sa/3.0/

Prime numbers.
"""

from bisect import bisect_left as bisect
from itertools import count
from math import sqrt


__primes = [
    2,
    3,
    5,
    7,
    11,
    13,
    17,
    19,
    23,
    29,
    31,
    37,
    41,
    43,
    47,
    53,
    59,
    61,
    67,
    71,
    73,
    79,
    83,
    89,
    97,
]


def nth_prime(n):
    global __primes  # noqa: PLW0603
    if not __primes:
        __primes = [2]

    if n <= len(__primes):
        return __primes[n - 1]

    i = __primes[-1]
    limit = 1 + bisect(__primes, int(sqrt(i)))
    while len(__primes) < n:
        i += 2
        while __primes[limit] ** 2 <= i:
            limit += 1
        if all(i % p for p in __primes[1:limit]):
            __primes.append(i)
    return __primes[-1]


def known_prime(n):
    if n <= __primes[-1]:
        i = bisect(__primes, n)
        return __primes[i] == n
    return False


def is_prime(n):
    n = abs(n)
    if n < 2:
        return False
    elif known_prime(n):
        return True
    else:
        return all(n % p for p in primes_upto(int(sqrt(n)) + 1))


def all_primes():
    for n in count(1):
        yield nth_prime(n)


def primes_upto(m, b=1):
    for p in all_primes():
        if p < b:
            continue
        if p > m:
            break
        yield p



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import threading
import time

from ..packetz import Packet, PacketzQueue
from .multi import Multi
from .row import BarRow


class BarBroker:
    def __init__(self, multi: Multi, fps: float = 60):
        self.multi = multi
        self.queue: PacketzQueue = PacketzQueue()
        self.reader: threading.Thread | None = None
        self.alive = False
        self.fps = fps

    def start(self):
        """Starts the completely isolated background rendering thread."""
        self.alive = True
        self.multi.start()

        self.reader = threading.Thread(target=self._read_queue, daemon=True)
        self.reader.start()

    def updated(self, row: BarRow):
        self.queue.send(data=row)

    def _read_queue(self):
        while self.alive:
            for value in self.queue.receive():
                if value is None:
                    continue
                match value:
                    case Packet(data=BarRow() as row):
                        self.multi.update_row(row.snap())
                    case _:
                        continue

            time.sleep(1 / self.fps)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
# by Copilot 2026-03-06
from __future__ import annotations

import pytest

from tatsu.util.strtools import pythonize_name, safe_name


@pytest.mark.parametrize(
    "name, plug, expected",
    [
        ("valid_name", "_", "valid_name"),
        ("123invalid", "_", "_123invalid"),
        ("name-with-dash", "_", "name_with_dash"),
        ("name with space", "_", "name_with_space"),
        ("class", "_", "class_"),
        ("name!", "_", "name_"),
        ("_leading", "_", "_leading"),
        ("123", "_", "_123"),
        ("if", "_", "if_"),
        ("type", "_", "type_"),
        ("name@domain", "_", "name_domain"),
        ("a-b_c.d", "_", "a_b_c_d"),
        ("1abc", "_", "_1abc"),
        ("abc1", "_", "abc1"),
        ("for", "_", "for_"),
        ("def", "_", "def_"),
    ],
)
def test_safe_name_valid_cases(name, plug, expected):
    assert safe_name(name, plug) == expected


@pytest.mark.parametrize(
    "name, plug",
    [
        ("", "_"),
        ("", "x"),
        ("name", ""),
    ],
)
def test_safe_name_invalid_inputs(name, plug):
    with pytest.raises(ValueError):
        safe_name(name, plug)


def test_safe_name_invalid_plug():
    with pytest.raises(ValueError):
        safe_name("name", "123")


@pytest.mark.parametrize(
    "name, expected",
    [
        ("", ""),
        ("someName", "some_name"),
        ("SomeName", "some_name"),
        ("XMLHttpRequest", "xml_http_request"),
        ("some_Name", "some__name"),
        ("name", "name"),
        ("NAME", "name"),
        ("some-name", "some_name"),
        ("123name", "_123name"),
        ("some name", "some_name"),
    ],
)
def test_pythonize_name(name, expected):
    assert pythonize_name(name) == expected


@pytest.mark.parametrize(
    "name, plug, expected",
    [
        ("naïve", "_", "naïve"),
        ("café", "_", "café"),
        ("Γαμμα", "_", "Γαμμα"),
        ("123Γ", "_", "_123Γ"),
    ],
)
def test_safe_name_unicode_cases(name, plug, expected):
    assert safe_name(name, plug) == expected


@pytest.mark.parametrize(
    "name, expected",
    [
        ("naïve", "naïve"),
        ("SomeΓαμμα", "someγαμμα"),
        ("ΓαμμαName", "γαμμα_name"),
    ],
)
def test_pythonize_name_unicode_cases(name, expected):
    assert pythonize_name(name) == expected



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .contexts.decorator import isname, leftrec, name, nomemo, rule, tatsumasu, token
from .objectmodel import NodeDataclassParams, nodedataclass as dataclass


__all__ = [
    'NodeDataclassParams',
    'dataclass',
    'isname',
    'name',
    'leftrec',
    'nomemo',
    'rule',
    'tatsumasu',
    'token',
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import json
import re
from typing import Any

from ..util.asjson import AsJSONMixin, asjson
from ..util.debugging import ERROR_print
from ..util.fromjson import JSONBase, fromjson
from ..util.misc import hash2str, new_id
from ..util.tty import tty_escape, tty_unescape
from .compact import compact_value, decompact_value


HASH_PATTERN = r'^\{"hash":"([\w\d]+)","data":(.*?)\}$'


class PacketError(Exception):
    pass


class CannotUnPacketError(PacketError):
    pass


class BadPacketError(PacketError):
    pass


class PacketHashError(BadPacketError):
    pass


class HasID(AsJSONMixin):
    id: str


class PacketLike(HasID):
    to: str | None
    data: Any


class WithID(HasID, JSONBase):
    id: str = "0xBAAD"

    def __new__(cls, *_args, **_kwargs):
        new = super().__new__(cls)
        new.id = new_id()
        return new


class Packet(PacketLike, WithID):
    to: str | None = None
    data: Any = None

    def __init__(self, /, *, to: str | None = None, data: Any = None):
        if to is not None:
            self.to = to
        if data is not None:
            self.data = data


def hashed(data: str) -> str:
    hash = hash2str(data)
    hashed = f'{{"hash":"{hash}","data":{data}}}'
    return hashed


def unhashed(hashed: str) -> str:
    if not (m := re.match(HASH_PATTERN, hashed)):
        raise ERROR_print(
            f"checksum missing: {hashed}",
            extype=BadPacketError,
        )
    try:
        hash, data = m.group(1, 2)
    except (AttributeError, IndexError, re.error) as e:
        raise ERROR_print(
            f"invalid packet: {hashed}",
            extype=BadPacketError,
        ) from e

    actual = hash2str(data)
    if hash != actual:
        raise ERROR_print(
            f"checksum mismatch: {hash} != {actual}",
            extype=PacketHashError,
        )
    return data


def pack(packet: PacketLike) -> str:
    value: Any = asjson(packet)
    value = compact_value(value)
    value = json.dumps(
        value,
        separators=(",", ":"),
        ensure_ascii=False,
    )
    value = class_escape(value)
    value = tty_escape(value)
    return hashed(value)


def unpack(hashed: str) -> PacketLike:
    try:
        value: Any = unhashed(hashed)
        value = tty_unescape(value)
        value = class_unescape(value)
        value = json.loads(value)
        value = decompact_value(value)
        return fromjson(value)
    except PacketError:
        raise
    except Exception as e:
        raise CannotUnPacketError(e) from e


def class_escape(s: str) -> str:
    return s.replace(r'"__class__":', '"@":')


def class_unescape(s: str) -> str:
    return s.replace(r'"@":', r'"__class__":')



#!/usr/bin/env -S uv run --script
# /// script
# dependencies = ["markdown-it-py"]
# ///
"""
Convert markdown to manpage-style plain text.

Usage:
    md2man [options] [input [output]]

If input is omitted, reads from stdin. If output is omitted, writes to stdout.
"""

import argparse
import sys
import textwrap
from pathlib import Path

from markdown_it import MarkdownIt
from markdown_it.tree import SyntaxTreeNode


def justify(text: str, width: int) -> str:
    """Justify wrapped text so every line (except the last) is exactly *width* wide."""
    lines = text.split("\n")
    out: list[str] = []
    for i, line in enumerate(lines):
        if i == len(lines) - 1 or not line.strip() or len(line) >= width:
            out.append(line)
        else:
            words = line.split()
            if len(words) <= 1:
                out.append(line)
                continue
            char_total = sum(len(w) for w in words)
            gap_count = len(words) - 1
            space_total = width - char_total
            base = space_total // gap_count
            extra = space_total % gap_count
            justified = words[0]
            for j, w in enumerate(words[1:], 1):
                justified += " " * (base + (1 if j <= extra else 0)) + w
            out.append(justified)
    return "\n".join(out)


def convert(md: str, width: int = 80, do_justify: bool = True) -> str:
    parser = MarkdownIt("default", {"linkify": False, "typographer": False})
    parser.enable(["table"])
    tokens = parser.parse(md)
    root = SyntaxTreeNode(tokens)

    out: list[str] = []
    list_stack: list[str] = []
    links: dict[str, str] = {}

    def render_inline(node: SyntaxTreeNode) -> str:
        parts: list[str] = []
        for child in node.children:
            t = child.type
            if t == "text":
                parts.append(child.content)
            elif t == "code_inline":
                parts.append(child.content)
            elif t in {"strong", "em", "s"}:
                parts.append(render_inline(child))
            elif t == "link":
                text = render_inline(child)
                href = child.attrs.get("href", "")
                if text and href:
                    links[text] = str(href)
                parts.append("[" + text + "]")
            elif t == "softbreak":
                parts.append(" ")
            elif t == "hardbreak":
                parts.append("\n")
            elif t == "image":
                parts.append(child.content or "")
            elif t == "html_inline":
                if not child.content.startswith("<!--"):
                    parts.append(child.content)
            else:
                parts.append(child.content or "")
        return "".join(parts)

    def blank():
        if not out:
            return
        if not out[-1]:
            return
        out.append("")

    def _do_justify(text: str, w: int) -> str:
        return justify(text, w) if do_justify else text

    def enter(node: SyntaxTreeNode):
        nonlocal list_stack
        t = node.type

        if t == "heading":
            _level = int(node.tag[1]) if node.tag.startswith("h") else 2
            inline_child = _first_inline(node)
            text = render_inline(inline_child) if inline_child else ""
            blank()
            out.append(text.upper())
            blank()

        if t == "bullet_list":
            indent = "  " + "  " * len(list_stack) + "* "
            list_stack.append(indent)
        elif t == "ordered_list":
            indent = "  " + "  " * len(list_stack) + "1. "
            list_stack.append(indent)
        elif t == "list_item":
            pass

        elif t == "paragraph":
            inline_child = _first_inline(node)
            if inline_child:
                text = render_inline(inline_child)
                if text.strip():
                    segments = text.split("\n")
                    if list_stack:
                        prefix = list_stack[-1]
                        avail = width - len(prefix)
                        if avail >= 20:
                            wrapped = []
                            for seg in segments:
                                w = textwrap.fill(seg, width=avail)
                                if w:
                                    w = _do_justify(w, avail)
                                    for line in w.split("\n"):
                                        wrapped.append(prefix + line)
                            text_out = "\n".join(wrapped)
                        else:
                            text_out = prefix + text
                    else:
                        wrapped = []
                        for seg in segments:
                            w = textwrap.fill(seg, width=width)
                            if w:
                                wrapped.append(_do_justify(w, width))
                        text_out = "\n".join(wrapped)
                    out.append(text_out)
            blank()

        elif t == "fence":
            content = node.content.rstrip("\n")
            if content:
                blank()
                for line in content.split("\n"):
                    out.append("    " + line)
                blank()

        elif t == "code_block":
            content = node.content.rstrip("\n")
            if content:
                blank()
                for line in content.split("\n"):
                    out.append("    " + line)
                blank()

        elif t == "hr":
            blank()
            out.append("-" * 72)
            blank()

        elif t == "table":
            blank()
            rows = []
            for child in node.children:
                if child.type in {"thead", "tbody"}:
                    for row in child.children:
                        if row.type == "tr":
                            cells = []
                            for cell in row.children:
                                if cell.type in {"th", "td"}:
                                    cells.append(render_inline(cell))
                            if cells:
                                rows.append("  " + " | ".join(cells))
            out.extend(rows)
            blank()

    def exit(node: SyntaxTreeNode):
        nonlocal list_stack
        t = node.type
        if t in {"bullet_list", "ordered_list"}:
            if list_stack:
                list_stack.pop()
            blank()
        elif t == "blockquote":
            blank()

    def walk(node: SyntaxTreeNode):
        enter(node)
        for child in node.children:
            walk(child)
        exit(node)

    walk(root)

    if links:
        out += ["", "LINKS", ""]
        for text, url in links.items():
            out.append(f"  [{text}]: {url}")
        out += [""]

    return "\n".join(out)


def _first_inline(node: SyntaxTreeNode) -> SyntaxTreeNode | None:
    for child in node.children:
        if child.type == "inline":
            return child
    return None


def main():
    parser = argparse.ArgumentParser(
        description="Convert markdown to plain text (manpage format).",
    )
    parser.add_argument("input", nargs="?", help="input file (stdin if omitted)")
    parser.add_argument("output", nargs="?", help="output file (stdout if omitted)")
    parser.add_argument(
        "-w", "--width", type=int, default=80, help="line width (default: 80)"
    )
    parser.add_argument(
        "--no-justify",
        action="store_false",
        dest="justify",
        help="disable line justification",
    )

    args = parser.parse_args()

    if args.input:
        md = Path(args.input).read_text()
    else:
        md = sys.stdin.read()

    result = convert(md, width=args.width, do_justify=args.justify)
    result = f"{result.rstrip()}\n"

    if args.output:
        Path(args.output).write_text(result)
    else:
        sys.stdout.write(result)


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations


def fromany(ss: Any) -> ParserStateStack:
    """
    Safely promotes an opaque state object to a ParserStateStack.
    This acts as the type-safe gateway for Engine contexts.
    """
    if not isinstance(ss, ParserStateStack):
        raise TypeError(
            f"Parser expected ParserStateStack, received {type(ss).__name__}"
        )
    return ss



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

# noqa # type: ignore # ruff: noqa
# type: ignore
# ty: ignore
# pyrefly: ignore
# pyright: ignore
# pyright: reportGeneralTypeIssues=false
# pyrefly: ignore-errors

from collections.abc import Iterable

from ..basic import Cut
from ..choice import Choice
from ..base import Box, Model, Rule
from ..syntax import Call, Sequence
from . import sccutils


__all__ = ['mark_left_recursion']


def _callable_rule_ids(exp: Model, rule_index: dict[str, int]) -> list[int]:
    if isinstance(exp, Call):
        target = rule_index.get(exp.name)
        return [target] if target is not None else []

    if isinstance(exp, Choice):
        result: list[int] = []
        for opt in exp.options:
            result.extend(_callable_rule_ids(opt, rule_index))
        return result

    if isinstance(exp, Sequence):
        result = []
        for item in exp.sequence:
            if isinstance(item, Cut):
                continue
            result.extend(_callable_rule_ids(item, rule_index))
            if not _is_nullable_safe(item):
                break
        return result

    if isinstance(exp, Box):
        return _callable_rule_ids(exp.exp, rule_index)

    return []


def _is_nullable_safe(exp: Model) -> bool:
    if isinstance(exp, Call):
        return False

    if isinstance(exp, Sequence):
        return all(_is_nullable_safe(item) for item in exp.sequence)

    if isinstance(exp, Choice):
        return any(_is_nullable_safe(opt) for opt in exp.options)

    return exp.is_nullable()


def _make_first_graph(
    rules: list[Rule], rule_index: dict[str, int]
) -> dict[str, set[str]]:
    graph: dict[str, set[str]] = {}
    for rule in rules:
        graph[rule.name] = {
            rules[i].name for i in _callable_rule_ids(rule.exp, rule_index)
        }

    all_vertices: set[str] = set(graph.keys())
    for deps in graph.values():
        all_vertices.update(deps)
    for v in all_vertices:
        graph.setdefault(v, set())

    return graph


def mark_left_recursion(rules: Iterable[Rule]) -> list[Rule]:
    rules = list(rules)
    rule_index = {rule.name: i for i, rule in enumerate(rules)}

    for rule in rules:
        rule.is_lrec = False
        rule.is_memo = not rule.no_memo

    graph = _make_first_graph(rules, rule_index)

    # pyrefly: ignore [bad-argument-type]
    sccs = list(sccutils.strongly_connected_components(set(graph.keys()), graph))  # type: ignore # pyright: ignore[reportArgumentType]  # ty:ignore[invalid-argument-type]

    for scc in sccs:
        if len(scc) > 1:
            for name in scc:
                rules[rule_index[name]].is_memo = False

            leaders = set(scc)
            for start in scc:
                for cycle in sccutils.find_cycles_in_scc(graph, scc, start):  # type: ignore # pyright: ignore[reportArgumentType]  # ty:ignore[invalid-argument-type]
                    leaders -= scc - set(cycle)
                    if not leaders:
                        break
                if not leaders:
                    break

            if not leaders:
                leaders = set(scc)

            leader_name = min(leaders)
            rules[rule_index[leader_name]].is_lrec = True

        elif len(scc) == 1:
            name = min(scc)
            if name in graph.get(name, set()):
                rule = rules[rule_index[name]]
                rule.is_lrec = True
                rule.is_memo = False

    return [rule for rule in rules if rule.is_lrec]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import io
import sys
from collections.abc import Generator
from contextlib import contextmanager
from datetime import UTC, datetime
from pathlib import Path


__all__ = ['logctx', 'iso_logpath', 'startscript']


def startscript() -> str:
    import __main__ as main

    if main.__package__:
        return main.__package__
    elif isinstance(getattr(main, '__file__', None), str):
        return Path(main.__file__).name
    else:
        return 'unknown'


def iso_timestamp() -> str:
    now = datetime.now(UTC)
    isostr = now.isoformat(timespec='seconds')
    return isostr.replace(':', '-').replace('+00-00', 'Z')


def iso_logpath(prefix: str, basedir: str = "log", suffix: str = '.log') -> Path:
    ts = iso_timestamp()
    logdir = Path(basedir)
    logdir.mkdir(parents=True, exist_ok=True)
    suffix = suffix if suffix.startswith('.') else '.' + suffix
    return logdir / f"{prefix}_{ts}{suffix}"


@contextmanager
def logctx(logpath: Path | None) -> Generator[io.TextIOBase, None, None]:
    if isinstance(logpath, Path):
        with logpath.open(mode="a", encoding="utf-8") as logfile:
            yield logfile
    else:
        yield sys.stderr  # type: ignore



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from pathlib import Path


PATH_TATSU_GRAMMAR = str(Path().cwd() / 'grammar/tatsu.ebnf')



#!/usr/bin/env python3
from __future__ import annotations

from .g2etool import (
    add_argparse_options,
    add_g2e_cmd,
    g2e_cmd,
    main,
    translate,
)


__all__ = [
    'main',
    'add_argparse_options',
    'translate',
    'g2e_cmd',
    'add_g2e_cmd',
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Any

from ..ngcodegen import modelgen, parsergen, pythongen
from ..peg import Grammar
from ..util.asjson import asjsons
from ..util.checkpygments import is_pygments_available
from .cfg import DEFAULT_PYGMENTS_STYLE, CLIConfig


def format_result(cfg: CLIConfig, result: Any) -> str:
    if cfg.model:
        return repr(result)
    return asjsons(result)


def colorize_output(
    payload: str,
    language: str,
    style: str = DEFAULT_PYGMENTS_STYLE,
) -> str:
    if not is_pygments_available():
        return payload

    from pygments import highlight
    from pygments.formatters import Terminal256Formatter
    from pygments.lexers import get_lexer_by_name

    lexer = get_lexer_by_name(language)
    fmt = Terminal256Formatter(style=style)
    return highlight(payload, lexer, fmt)


def render_grammar(
    model: Grammar,
    cfg: CLIConfig,
    *,
    name: str | None = None,
) -> str:
    """Render a Grammar in the selected mode.

    Sets *name* to the output basename (without extension) when -o is a directory.
    """
    if cfg.optimized:
        model = model.optimized()

    _ = name
    if cfg.generate_parser:
        result = pythongen(model, parser_name=cfg.name)
    elif cfg.model:
        result = repr(model)
    elif cfg.railroads:
        result = model.railroads()
    elif cfg.pretty:
        result = model.pretty()
    elif cfg.pretty_lean:
        result = model.pretty_lean()
    elif cfg.object_model:
        result = modelgen(model, name=cfg.name)
    elif cfg.parser_model:
        result = parsergen(model, name=cfg.name)
    else:
        result = model.asjsons()
    return result



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from types import TracebackType
from typing import Any

from ..ctx import Ctx


class ContextBase:
    def __init__(self, ctx: Ctx):
        self.ctx = ctx
        self.expected: list[str] = []

    def expecting(self, *tokens: str) -> None:
        self.expected.extend(tokens)

    def expectedexcept(self, ctx: Ctx) -> Exception:
        msg = 'Failed'
        if self.expected:
            msg = f'Expected one of: {' '.join(self.expected)}'
        return ctx.newexcept(msg)

    def parse(self, ctx: Ctx) -> Any:
        _ = ctx
        return None

    def __enter__(self) -> ContextBase:
        return self

    def __exit__(
        self,
        etype: type[BaseException] | None,
        evalue: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        if etype is not None:
            return False

        self.parse(self.ctx)
        return True



#!/usr/bin/env python3
#
#  CAVEAT UTILITOR
#
#  This file was automatically generated by 竜TatSu v5.23.1b1
#
#     https://pypi.python.org/pypi/tatsu/
#
#  Any changes you make to it will be overwritten the next time
#  the file is generated.
#
# noqa # type: ignore # ruff: noqa
# copyright: ignore
# fmt: off

from __future__ import annotations

from typing import Any

from tatsu import decorators as tatsu
from tatsu.contexts import Ctx
from tatsu.config import ParserConfig
from tatsu.parsing import Parser
from tatsu.input.buffer import Buffer
from tatsu.input.textlines import TextLines

KEYWORDS = ()


class TatSuBootstrapText(TextLines):
    def __init__(
        self,
        text,
        /,
        config: ParserConfig | None = None,
        **settings,
    ) -> None:
        config = ParserConfig.new(
            config=config,
            name='TatSu',
            whitespace=None,
            nameguard=None,
            ignorecase=False,
            namechars='',
            parseinfo=True,
            comments=r'(?ms)[(][*]\s*(.*?)\s*[*][)]|/[*]\s*(.*?)\s*[*]/',
            eol_comments=r'(?ms)(?:#|//)(.*?)$',
            keywords=KEYWORDS,
            start='start',
        )
        assert isinstance(config, ParserConfig)
        config = config.override(**settings)

        super().__init__(text, config=config)

TatSuBootstrapTokenizer = TatSuBootstrapText


class TatSuBootstrapBuffer(Buffer):  # NOTE: backwards compatibility
    def __init__(
        self,
        text,
        /,
        config: ParserConfig | None = None,
        **settings,
    ) -> None:
        config = ParserConfig.new(
            config=config,
            name='TatSu',
            whitespace=None,
            nameguard=None,
            ignorecase=False,
            namechars='',
            parseinfo=True,
            comments=r'(?ms)[(][*]\s*(.*?)\s*[*][)]|/[*]\s*(.*?)\s*[*]/',
            eol_comments=r'(?ms)(?:#|//)(.*?)$',
            keywords=KEYWORDS,
            start='start',
        )
        assert isinstance(config, ParserConfig)
        config = config.override(**settings)

        super().__init__(text, config=config)


class TatSuBootstrapParser(Parser):
    def __init__(self, /, config: ParserConfig | None = None, **settings):
        config = ParserConfig.new(config, **settings)
        rulessource = TatSuBootstrapRules()
        assert isinstance(config, ParserConfig)

        super().__init__(rulessource, config=config)


class TatSuBootstrapRules:
    def __init__(self, /, config: ParserConfig | None = None, **settings):
        config = ParserConfig.new(
            config=config,
            name='TatSu',
            whitespace=None,
            nameguard=None,
            ignorecase=False,
            namechars='',
            parseinfo=True,
            comments=r'(?ms)[(][*]\s*(.*?)\s*[*][)]|/[*]\s*(.*?)\s*[*]/',
            eol_comments=r'(?ms)(?:#|//)(.*?)$',
            keywords=KEYWORDS,
            start='start',
        )
        assert isinstance(config, ParserConfig)
        config = config.override(**settings)
        self._config = config


    @tatsu.rule
    def start(self, ctx: Ctx) -> Any:
        self.grammar(ctx)

    @tatsu.rule('Grammar')
    def grammar(self, ctx: Ctx) -> Any:
        ctx.define(
            ['directives', 'keywords', 'rules', 'title'],
            ['directives', 'keywords', 'rules'],
        )
        with ctx.nameset('title'):
            ctx.constant('TATSU')
        with ctx.loopopt() as cl:
            @cl.exp
            def _(ctx: Ctx) -> Any:
                with ctx.choice() as α:
                    α.expecting('@@', '@@keyword', '→directive', '→keyword')

                    @α.option
                    def _(ctx: Ctx) -> Any:
                        with ctx.nameadd('directives'):
                            self.directive(ctx)
                    @α.option
                    def _(ctx: Ctx) -> Any:
                        with ctx.nameadd('keywords'):
                            self.keyword(ctx)
        with ctx.nameadd('rules'):
            self.rule(ctx)
        with ctx.loopopt() as clb:
            @clb.exp
            def _(ctx: Ctx) -> Any:
                with ctx.choice() as α:
                    α.expecting('<', '=|::=|:=?', '@@keyword', '→decorator', '→keyword', '→rule')

                    @α.option
                    def _(ctx: Ctx) -> Any:
                        with ctx.nameadd('rules'):
                            self.rule(ctx)
                    @α.option
                    def _(ctx: Ctx) -> Any:
                        with ctx.nameadd('keywords'):
                            self.keyword(ctx)
        ctx.eofcheck()

    @tatsu.rule
    def directive(self, ctx: Ctx) -> Any:
        ctx.define(['name', 'value'], [])
        ctx.token('@@')
        with ctx.ifnot_():
            ctx.token('keyword')
        ctx.cut()
        with ctx.group():
            with ctx.choice() as α:
                α.expecting(
                  'comments',
                  'eol_comments',
                  'grammar',
                  'ignorecase',
                  'left_recursion',
                  'memoization',
                  'namechars',
                  'nameguard',
                  'parseinfo',
                  'whitespace'
                )

                @α.option
                def _(ctx: Ctx) -> Any:
                    ctx.define(['name', 'value'], [])
                    with ctx.nameset('name'):
                        with ctx.group():
                            with ctx.choice() as β:
                                β.expecting('comments', 'eol_comments')

                                @β.option
                                def _(ctx: Ctx) -> Any:
                                    ctx.token('comments')
                                @β.option
                                def _(ctx: Ctx) -> Any:
                                    ctx.token('eol_comments')
                    ctx.cut()
                    ctx.token('::')
                    ctx.cut()
                    with ctx.nameset('value'):
                        self.regex(ctx)
                @α.option
                def _(ctx: Ctx) -> Any:
                    ctx.define(['name', 'value'], [])
                    with ctx.nameset('name'):
                        ctx.token('whitespace')
                    ctx.cut()
                    ctx.token('::')
                    ctx.cut()
                    with ctx.nameset('value'):
                        with ctx.group():
                            with ctx.choice() as β:
                                β.expecting('False', 'None', '→regex', '→string')

                                @β.option
                                def _(ctx: Ctx) -> Any:
                                    self.regex(ctx)
                                @β.option
                                def _(ctx: Ctx) -> Any:
                                    self.string(ctx)
                                @β.option
                                def _(ctx: Ctx) -> Any:
                                    ctx.token('None')
                                @β.option
                                def _(ctx: Ctx) -> Any:
                                    ctx.token('False')
                                @β.option
                                def _(ctx: Ctx) -> Any:
                                    ctx.constant(None)
                @α.option
                def _(ctx: Ctx) -> Any:
                    ctx.define(['name', 'value'], [])
                    with ctx.nameset('name'):
                        with ctx.group():
                            with ctx.choice() as β:
                                β.expecting('ignorecase', 'left_recursion', 'memoization', 'nameguard', 'parseinfo')

                                @β.option
                                def _(ctx: Ctx) -> Any:
                                    ctx.token('nameguard')
                                @β.option
                                def _(ctx: Ctx) -> Any:
                                    ctx.token('ignorecase')
                                @β.option
                                def _(ctx: Ctx) -> Any:
                                    ctx.token('left_recursion')
                                @β.option
                                def _(ctx: Ctx) -> Any:
                                    ctx.token('parseinfo')
                                @β.option
                                def _(ctx: Ctx) -> Any:
                                    ctx.token('memoization')
                    ctx.cut()
                    with ctx.group():
                        with ctx.choice() as β:
                            β.expecting('::')

                            @β.option
                            def _(ctx: Ctx) -> Any:
                                ctx.define(['value'], [])
                                ctx.token('::')
                                ctx.cut()
                                with ctx.nameset('value'):
                                    self.boolean(ctx)
                            @β.option
                            def _(ctx: Ctx) -> Any:
                                with ctx.nameset('value'):
                                    ctx.constant(True)
                @α.option
                def _(ctx: Ctx) -> Any:
                    ctx.define(['name', 'value'], [])
                    with ctx.nameset('name'):
                        ctx.token('grammar')
                    ctx.cut()
                    ctx.token('::')
                    ctx.cut()
                    with ctx.nameset('value'):
                        self.word(ctx)
                @α.option
                def _(ctx: Ctx) -> Any:
                    ctx.define(['name', 'value'], [])
                    with ctx.nameset('name'):
                        ctx.token('namechars')
                    ctx.cut()
                    ctx.token('::')
                    ctx.cut()
                    with ctx.nameset('value'):
                        self.string(ctx)
        ctx.cut()

    @tatsu.rule
    def keywords(self, ctx: Ctx) -> Any:
        with ctx.loopplus() as cl:
            @cl.exp
            def _(ctx: Ctx) -> Any:
                self.keyword(ctx)

    @tatsu.rule
    def keyword(self, ctx: Ctx) -> Any:
        ctx.token('@@keyword')
        ctx.cut()
        ctx.token('::')
        with ctx.group():
            with ctx.choice() as α:
                α.expecting('(', '→string', '→word')

                @α.option
                def _(ctx: Ctx) -> Any:
                    ctx.token('(')
                    ctx.cut()
                    with ctx.result():
                        with ctx.loopplus() as cl:
                            @cl.exp
                            def _(ctx: Ctx) -> Any:
                                with ctx.choice() as β:
                                    β.expecting('→string', '→word')

                                    @β.option
                                    def _(ctx: Ctx) -> Any:
                                        self.word(ctx)
                                    @β.option
                                    def _(ctx: Ctx) -> Any:
                                        self.string(ctx)
                    ctx.token(')')
                @α.option
                def _(ctx: Ctx) -> Any:
                    with ctx.loopplus() as clb:
                        @clb.exp
                        def _(ctx: Ctx) -> Any:
                            with ctx.resultadd():
                                with ctx.group():
                                    with ctx.choice() as β:
                                        β.expecting('→string', '→word')

                                        @β.option
                                        def _(ctx: Ctx) -> Any:
                                            self.word(ctx)
                                        @β.option
                                        def _(ctx: Ctx) -> Any:
                                            self.string(ctx)
                            with ctx.ifnot_():
                                with ctx.group():
                                    with ctx.choice() as β:
                                        β.expecting(':', '=')

                                        @β.option
                                        def _(ctx: Ctx) -> Any:
                                            ctx.token(':')
                                        @β.option
                                        def _(ctx: Ctx) -> Any:
                                            ctx.token('=')

    @tatsu.rule
    def params(self, ctx: Ctx) -> Any:
        with ctx.resultadd():
            self.first_param(ctx)
        with ctx.loopopt() as cl:
            @cl.exp
            def _(ctx: Ctx) -> Any:
                ctx.token(',')
                with ctx.resultadd():
                    self.literal(ctx)
                with ctx.ifnot_():
                    ctx.token('=')
                ctx.cut()

    @tatsu.rule
    def first_param(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              '[_\\w][_\\w\\d]*(?:::[_\\w][_\\w\\d]*)+',
              '→boolean',
              '→float',
              '→hex',
              '→int',
              '→literal',
              '→none',
              '→path',
              '→raw_string',
              '→value',
              '→word'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                self.path(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.literal(ctx)

    @tatsu.rule
    def kwparams(self, ctx: Ctx) -> Any:
        with ctx.gatherplus() as g:
            @g.sep
            def _(ctx: Ctx) -> Any:
                ctx.token(',')

            @g.exp
            def _(ctx: Ctx) -> Any:
                self.pair(ctx)

    @tatsu.rule
    def the_params_at_last(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              '→first_param',
              '→kwparams',
              '→literal',
              '→pair',
              '→params',
              '→path',
              '→word'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                with ctx.nameset('kwparams'):
                    self.kwparams(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.define(['kwparams', 'params'], [])
                with ctx.nameset('params'):
                    self.params(ctx)
                ctx.token(',')
                ctx.cut()
                with ctx.nameset('kwparams'):
                    self.kwparams(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                with ctx.nameset('params'):
                    self.params(ctx)

    @tatsu.rule
    def paramdef(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting('(', '::', '[')

            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.define(['kwparams', 'params'], [])
                ctx.token('[')
                ctx.cut()
                with ctx.choice() as β:
                    β.expecting(
                      '→first_param',
                      '→kwparams',
                      '→literal',
                      '→pair',
                      '→params',
                      '→path',
                      '→word'
                    )

                    @β.option
                    def _(ctx: Ctx) -> Any:
                        with ctx.nameset('kwparams'):
                            self.kwparams(ctx)
                    @β.option
                    def _(ctx: Ctx) -> Any:
                        ctx.define(['kwparams', 'params'], [])
                        with ctx.nameset('params'):
                            self.params(ctx)
                        ctx.token(',')
                        ctx.cut()
                        with ctx.nameset('kwparams'):
                            self.kwparams(ctx)
                    @β.option
                    def _(ctx: Ctx) -> Any:
                        with ctx.nameset('params'):
                            self.params(ctx)
                ctx.token(']')
            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.define(['kwparams', 'params'], [])
                ctx.token('(')
                ctx.cut()
                with ctx.choice() as β:
                    β.expecting(
                      '→first_param',
                      '→kwparams',
                      '→literal',
                      '→pair',
                      '→params',
                      '→path',
                      '→word'
                    )

                    @β.option
                    def _(ctx: Ctx) -> Any:
                        with ctx.nameset('kwparams'):
                            self.kwparams(ctx)
                    @β.option
                    def _(ctx: Ctx) -> Any:
                        ctx.define(['kwparams', 'params'], [])
                        with ctx.nameset('params'):
                            self.params(ctx)
                        ctx.token(',')
                        ctx.cut()
                        with ctx.nameset('kwparams'):
                            self.kwparams(ctx)
                    @β.option
                    def _(ctx: Ctx) -> Any:
                        with ctx.nameset('params'):
                            self.params(ctx)
                ctx.token(')')
            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.define(['params'], [])
                ctx.token('::')
                ctx.cut()
                with ctx.nameset('params'):
                    self.params(ctx)

    @tatsu.rule('Rule')
    def rule(self, ctx: Ctx) -> Any:
        ctx.define(['base', 'decorators', 'exp', 'kwparams', 'name', 'params'], [])
        with ctx.nameset('decorators'):
            with ctx.loopopt() as cl:
                @cl.exp
                def _(ctx: Ctx) -> Any:
                    self.decorator(ctx)
        with ctx.nameset('name'):
            ctx.matchname()
        ctx.cut()
        with ctx.optional():
            with ctx.choice() as α:
                α.expecting('(', '::', '[')

                @α.option
                def _(ctx: Ctx) -> Any:
                    ctx.define(['kwparams', 'params'], [])
                    ctx.token('[')
                    ctx.cut()
                    with ctx.choice() as β:
                        β.expecting(
                          '→first_param',
                          '→kwparams',
                          '→literal',
                          '→pair',
                          '→params',
                          '→path',
                          '→word'
                        )

                        @β.option
                        def _(ctx: Ctx) -> Any:
                            with ctx.nameset('kwparams'):
                                self.kwparams(ctx)
                        @β.option
                        def _(ctx: Ctx) -> Any:
                            ctx.define(['kwparams', 'params'], [])
                            with ctx.nameset('params'):
                                self.params(ctx)
                            ctx.token(',')
                            ctx.cut()
                            with ctx.nameset('kwparams'):
                                self.kwparams(ctx)
                        @β.option
                        def _(ctx: Ctx) -> Any:
                            with ctx.nameset('params'):
                                self.params(ctx)
                    ctx.token(']')
                @α.option
                def _(ctx: Ctx) -> Any:
                    ctx.define(['kwparams', 'params'], [])
                    ctx.token('(')
                    ctx.cut()
                    with ctx.choice() as β:
                        β.expecting(
                          '→first_param',
                          '→kwparams',
                          '→literal',
                          '→pair',
                          '→params',
                          '→path',
                          '→word'
                        )

                        @β.option
                        def _(ctx: Ctx) -> Any:
                            with ctx.nameset('kwparams'):
                                self.kwparams(ctx)
                        @β.option
                        def _(ctx: Ctx) -> Any:
                            ctx.define(['kwparams', 'params'], [])
                            with ctx.nameset('params'):
                                self.params(ctx)
                            ctx.token(',')
                            ctx.cut()
                            with ctx.nameset('kwparams'):
                                self.kwparams(ctx)
                        @β.option
                        def _(ctx: Ctx) -> Any:
                            with ctx.nameset('params'):
                                self.params(ctx)
                    ctx.token(')')
                @α.option
                def _(ctx: Ctx) -> Any:
                    ctx.define(['params'], [])
                    ctx.token('::')
                    ctx.cut()
                    with ctx.nameset('params'):
                        self.params(ctx)
        with ctx.optional():
            ctx.define(['base'], [])
            ctx.token('<')
            ctx.cut()
            with ctx.nameset('base'):
                self.known_name(ctx)
        ctx.void()
        ctx.pattern(r'=|::=|:=?')
        ctx.cut()
        with ctx.nameset('exp'):
            self.expre(ctx)
        self.ENDRULE(ctx)
        ctx.cut()

    @tatsu.rule
    @tatsu.token
    def ENDRULE(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(';', '→BLANK', '→DEDENT', '→EOL')

            @α.option
            def _(ctx: Ctx) -> Any:
                self.DEDENT(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.BLANK(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.token(';')
            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.eofcheck()

    @tatsu.rule
    @tatsu.token
    def DEDENT(self, ctx: Ctx) -> Any:
        self.EOL(ctx)
        with ctx.if_():
            ctx.pattern(r'^\S')

    @tatsu.rule
    @tatsu.token
    def BLANK(self, ctx: Ctx) -> Any:
        self.EOL(ctx)
        self.EOL(ctx)

    @tatsu.rule
    @tatsu.token
    def EOL(self, ctx: Ctx) -> Any:
        ctx.pattern(r'(?m)[ \t]*$')
        ctx.pattern(r'(?m)(?:\r?\n|\r)')

    @tatsu.rule
    def decorator(self, ctx: Ctx) -> Any:
        ctx.token('@')
        with ctx.ifnot_():
            ctx.token('@')
        ctx.cut()
        with ctx.result():
            ctx.pattern(r'(override|name|isname|nomemo|nostak)\b')

    @tatsu.rule
    def pair(self, ctx: Ctx) -> Any:
        with ctx.resultadd():
            self.word(ctx)
        ctx.token('=')
        ctx.cut()
        with ctx.resultadd():
            self.literal(ctx)

    @tatsu.rule
    def expre(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              ';',
              '→BLANK',
              '→DEDENT',
              '→ENDRULE',
              '→choice',
              '→element',
              '→option',
              '→sequence',
              '|'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                self.choice(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.sequence(ctx)

    @tatsu.rule('Choice')
    def choice(self, ctx: Ctx) -> Any:
        with ctx.optional():
            ctx.token('|')
            ctx.cut()
        with ctx.resultadd():
            self.option(ctx)
        with ctx.loopplus() as cl:
            @cl.exp
            def _(ctx: Ctx) -> Any:
                ctx.token('|')
                ctx.cut()
                with ctx.resultadd():
                    self.option(ctx)

    @tatsu.rule('Option')
    def option(self, ctx: Ctx) -> Any:
        self.sequence(ctx)

    @tatsu.rule('Sequence')
    def sequence(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              ';',
              '→BLANK',
              '→DEDENT',
              '→ENDRULE',
              '→EOL',
              '→element',
              '→meta',
              '→named',
              '→override',
              '→rule_include',
              '→term'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                with ctx.if_():
                    with ctx.group():
                        self.element(ctx)
                        ctx.token(',')
                with ctx.gatherplus() as g:
                    @g.sep
                    def _(ctx: Ctx) -> Any:
                        ctx.token(',')

                    @g.exp
                    def _(ctx: Ctx) -> Any:
                        self.element(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                with ctx.loopplus() as cl:
                    @cl.exp
                    def _(ctx: Ctx) -> Any:
                        with ctx.ifnot_():
                            self.ENDRULE(ctx)
                        self.element(ctx)

    @tatsu.rule
    def element(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              '>',
              '@(name|int|uint|float|bool)\\b',
              '→atom',
              '→closure',
              '→cut',
              '→cut_deprecated',
              '→empty_closure',
              '→gather',
              '→join',
              '→left_join',
              '→lookahead',
              '→meta',
              '→named',
              '→named_list',
              '→named_single',
              '→negative_lookahead',
              '→optional',
              '→override',
              '→override_list',
              '→override_single',
              '→positive_closure',
              '→right_join',
              '→rule_include',
              '→skip_to',
              '→term',
              '→void'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                self.override(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.meta(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.named(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.term(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.rule_include(ctx)

    @tatsu.rule('RuleInclude')
    def rule_include(self, ctx: Ctx) -> Any:
        ctx.token('>')
        ctx.cut()
        with ctx.result():
            self.known_name(ctx)

    @tatsu.rule
    def named(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting('[:=]', '\\+[:=]', '→named_list', '→named_single')

            @α.option
            def _(ctx: Ctx) -> Any:
                self.named_list(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.named_single(ctx)

    @tatsu.rule('NamedList')
    def named_list(self, ctx: Ctx) -> Any:
        ctx.define(['exp', 'name'], [])
        with ctx.nameset('name'):
            ctx.matchname()
        ctx.pattern(r'\+[:=]')
        ctx.cut()
        with ctx.nameset('exp'):
            self.term(ctx)

    @tatsu.rule('Named')
    def named_single(self, ctx: Ctx) -> Any:
        ctx.define(['exp', 'name'], [])
        with ctx.nameset('name'):
            ctx.matchname()
        ctx.pattern(r'[:=]')
        ctx.cut()
        with ctx.nameset('exp'):
            self.term(ctx)

    @tatsu.rule
    def override(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting('=|@:', '\\+=|@\\+:', '→override_list', '→override_single')

            @α.option
            def _(ctx: Ctx) -> Any:
                self.override_list(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.override_single(ctx)

    @tatsu.rule('OverrideList')
    def override_list(self, ctx: Ctx) -> Any:
        ctx.pattern(r'\+=|@\+:')
        ctx.cut()
        with ctx.result():
            self.term(ctx)

    @tatsu.rule('Override')
    def override_single(self, ctx: Ctx) -> Any:
        ctx.pattern(r'=|@:')
        ctx.cut()
        with ctx.result():
            self.term(ctx)

    @tatsu.rule
    def term(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              '!',
              '&',
              '()',
              '(?:',
              '->',
              '>>',
              '[',
              '→alert',
              '→atom',
              '→call',
              '→closure',
              '→constant',
              '→cut',
              '→cut_deprecated',
              '→dot',
              '→empty_closure',
              '→eof',
              '→eol',
              '→gather',
              '→group',
              '→join',
              '→left_join',
              '→lookahead',
              '→meta',
              '→negative_lookahead',
              '→optional',
              '→pattern',
              '→positive_closure',
              '→right_join',
              '→skip',
              '→skip_to',
              '→token',
              '→void',
              '{',
              '{}',
              '~'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                self.gather(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.join(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.left_join(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.right_join(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.empty_closure(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.positive_closure(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.closure(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.optional(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.atom(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.void(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.skip_to(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.lookahead(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.negative_lookahead(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.cut(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.cut_deprecated(ctx)

    @tatsu.rule('Group')
    def group(self, ctx: Ctx) -> Any:
        with ctx.ifnot_():
            ctx.token('(?:')
        ctx.token('(')
        ctx.cut()
        with ctx.result():
            self.expre(ctx)
        ctx.token(')')
        ctx.cut()

    @tatsu.rule('SkipGroup')
    def skip(self, ctx: Ctx) -> Any:
        ctx.token('(?:')
        ctx.cut()
        with ctx.result():
            self.expre(ctx)
        ctx.token(')')
        ctx.cut()

    @tatsu.rule
    def gather(self, ctx: Ctx) -> Any:
        with ctx.if_():
            with ctx.group():
                self.atom(ctx)
                ctx.token('.{')
        ctx.cut()
        with ctx.group():
            with ctx.choice() as α:
                α.expecting('→normal_gather', '→positive_gather')

                @α.option
                def _(ctx: Ctx) -> Any:
                    self.positive_gather(ctx)
                @α.option
                def _(ctx: Ctx) -> Any:
                    self.normal_gather(ctx)

    @tatsu.rule('PositiveGather')
    def positive_gather(self, ctx: Ctx) -> Any:
        ctx.define(['exp', 'sep'], [])
        with ctx.nameset('sep'):
            self.atom(ctx)
        ctx.token('.{')
        with ctx.nameset('exp'):
            self.expre(ctx)
        ctx.token('}')
        with ctx.ifnot_():
            ctx.pattern(r'\+=')
        ctx.pattern(r'[+-]')
        ctx.cut()

    @tatsu.rule('Gather')
    def normal_gather(self, ctx: Ctx) -> Any:
        ctx.define(['exp', 'sep'], [])
        with ctx.nameset('sep'):
            self.atom(ctx)
        ctx.token('.{')
        ctx.cut()
        with ctx.nameset('exp'):
            self.expre(ctx)
        ctx.token('}')
        with ctx.optional():
            ctx.token('*')
            ctx.cut()
        ctx.cut()

    @tatsu.rule
    def join(self, ctx: Ctx) -> Any:
        with ctx.if_():
            with ctx.group():
                self.atom(ctx)
                ctx.token('%{')
        ctx.cut()
        with ctx.group():
            with ctx.choice() as α:
                α.expecting('→normal_join', '→positive_join')

                @α.option
                def _(ctx: Ctx) -> Any:
                    self.positive_join(ctx)
                @α.option
                def _(ctx: Ctx) -> Any:
                    self.normal_join(ctx)

    @tatsu.rule('PositiveJoin')
    def positive_join(self, ctx: Ctx) -> Any:
        ctx.define(['exp', 'sep'], [])
        with ctx.nameset('sep'):
            self.atom(ctx)
        ctx.token('%{')
        with ctx.nameset('exp'):
            self.expre(ctx)
        ctx.token('}')
        with ctx.ifnot_():
            ctx.pattern(r'\+=')
        ctx.pattern(r'[+-]')
        ctx.cut()

    @tatsu.rule('Join')
    def normal_join(self, ctx: Ctx) -> Any:
        ctx.define(['exp', 'sep'], [])
        with ctx.nameset('sep'):
            self.atom(ctx)
        ctx.token('%{')
        ctx.cut()
        with ctx.nameset('exp'):
            self.expre(ctx)
        ctx.token('}')
        with ctx.optional():
            ctx.token('*')
            ctx.cut()
        ctx.cut()

    @tatsu.rule('LeftJoin')
    def left_join(self, ctx: Ctx) -> Any:
        ctx.define(['exp', 'sep'], [])
        with ctx.nameset('sep'):
            self.atom(ctx)
        ctx.token('<{')
        ctx.cut()
        with ctx.nameset('exp'):
            self.expre(ctx)
        ctx.token('}')
        with ctx.ifnot_():
            ctx.pattern(r'\+=')
        ctx.pattern(r'[+-]')
        ctx.cut()

    @tatsu.rule('RightJoin')
    def right_join(self, ctx: Ctx) -> Any:
        ctx.define(['exp', 'sep'], [])
        with ctx.nameset('sep'):
            self.atom(ctx)
        ctx.token('>{')
        ctx.cut()
        with ctx.nameset('exp'):
            self.expre(ctx)
        ctx.token('}')
        with ctx.ifnot_():
            ctx.pattern(r'\+=')
        ctx.pattern(r'[+-]')
        ctx.cut()

    @tatsu.rule('PositiveClosure')
    def positive_closure(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              '(?:',
              '→alert',
              '→atom',
              '→call',
              '→constant',
              '→dot',
              '→eof',
              '→eol',
              '→group',
              '→meta',
              '→pattern',
              '→skip',
              '→token',
              '{'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.token('{')
                with ctx.result():
                    self.expre(ctx)
                ctx.token('}')
                with ctx.ifnot_():
                    ctx.pattern(r'\+=')
                ctx.pattern(r'[+-]')
                ctx.cut()
            @α.option
            def _(ctx: Ctx) -> Any:
                with ctx.result():
                    self.atom(ctx)
                with ctx.ifnot_():
                    ctx.pattern(r'\+=')
                ctx.pattern(r'[+]')
                ctx.cut()

    @tatsu.rule('Closure')
    def closure(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              '(?:',
              '→alert',
              '→atom',
              '→call',
              '→constant',
              '→dot',
              '→eof',
              '→eol',
              '→group',
              '→meta',
              '→pattern',
              '→skip',
              '→token',
              '{'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.token('{')
                with ctx.result():
                    self.expre(ctx)
                ctx.token('}')
                with ctx.optional():
                    ctx.token('*')
                ctx.cut()
            @α.option
            def _(ctx: Ctx) -> Any:
                with ctx.result():
                    self.atom(ctx)
                ctx.token('*')
                ctx.cut()

    @tatsu.rule('EmptyClosure')
    def empty_closure(self, ctx: Ctx) -> Any:
        ctx.token('{}')
        ctx.cut()
        with ctx.result():
            ctx.void()

    @tatsu.rule('Optional')
    def optional(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              '(?:',
              '[',
              '→alert',
              '→atom',
              '→call',
              '→constant',
              '→dot',
              '→eof',
              '→eol',
              '→group',
              '→meta',
              '→pattern',
              '→skip',
              '→token'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.token('[')
                ctx.cut()
                with ctx.result():
                    self.expre(ctx)
                ctx.token(']')
                ctx.cut()
            @α.option
            def _(ctx: Ctx) -> Any:
                with ctx.result():
                    self.atom(ctx)
                with ctx.ifnot_():
                    with ctx.group():
                        with ctx.choice() as β:
                            β.expecting('?"', "?'", '?/')

                            @β.option
                            def _(ctx: Ctx) -> Any:
                                ctx.token('?"')
                            @β.option
                            def _(ctx: Ctx) -> Any:
                                ctx.token("?'")
                            @β.option
                            def _(ctx: Ctx) -> Any:
                                ctx.token('?/')
                ctx.token('?')
                ctx.cut()

    @tatsu.rule('Lookahead')
    def lookahead(self, ctx: Ctx) -> Any:
        ctx.token('&')
        ctx.cut()
        with ctx.result():
            self.term(ctx)

    @tatsu.rule('NegativeLookahead')
    def negative_lookahead(self, ctx: Ctx) -> Any:
        ctx.token('!')
        ctx.cut()
        with ctx.result():
            self.term(ctx)

    @tatsu.rule('SkipTo')
    def skip_to(self, ctx: Ctx) -> Any:
        ctx.token('->')
        ctx.cut()
        with ctx.result():
            self.term(ctx)

    @tatsu.rule
    def atom(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              '$',
              '$->',
              '(?:',
              '/./',
              '@(name|int|uint|float|bool)\\b',
              '\\^+',
              '`',
              '→alert',
              '→call',
              '→constant',
              '→dot',
              '→eof',
              '→eol',
              '→group',
              '→meta',
              '→pattern',
              '→raw_string',
              '→regex',
              '→skip',
              '→string',
              '→token'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                self.meta(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.token(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.call(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.dot(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.pattern(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.skip(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.group(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.eol(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.eof(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.alert(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.constant(ctx)

    @tatsu.rule('Meta')
    def meta(self, ctx: Ctx) -> Any:
        ctx.pattern(r'@(name|int|uint|float|bool)\b')

    @tatsu.rule('Call')
    def call(self, ctx: Ctx) -> Any:
        ctx.matchname()

    @tatsu.rule('Void')
    def void(self, ctx: Ctx) -> Any:
        ctx.token('()')
        ctx.cut()

    @tatsu.rule('Fail')
    def fail(self, ctx: Ctx) -> Any:
        ctx.token('!()')
        ctx.cut()

    @tatsu.rule('Cut')
    def cut(self, ctx: Ctx) -> Any:
        ctx.token('~')
        ctx.cut()

    @tatsu.rule('Cut')
    def cut_deprecated(self, ctx: Ctx) -> Any:
        ctx.token('>>')
        ctx.cut()

    @tatsu.rule
    def known_name(self, ctx: Ctx) -> Any:
        ctx.matchname()
        ctx.cut()

    @tatsu.rule
    def name(self, ctx: Ctx) -> Any:
        self.word(ctx)

    @tatsu.rule('Constant')
    def constant(self, ctx: Ctx) -> Any:
        with ctx.if_():
            ctx.token('`')
        with ctx.group():
            with ctx.choice() as α:
                α.expecting('(?ms)```((?:.|\\n)*?)```', '`', '`(.*?)`')

                @α.option
                def _(ctx: Ctx) -> Any:
                    ctx.pattern(r'(?ms)```((?:.|\n)*?)```')
                @α.option
                def _(ctx: Ctx) -> Any:
                    ctx.token('`')
                    with ctx.result():
                        self.literal(ctx)
                    ctx.token('`')
                @α.option
                def _(ctx: Ctx) -> Any:
                    ctx.pattern(r'`(.*?)`')

    @tatsu.rule('Alert')
    def alert(self, ctx: Ctx) -> Any:
        ctx.define(['level', 'message'], [])
        with ctx.nameset('level'):
            ctx.pattern(r'\^+')
        with ctx.nameset('message'):
            self.constant(ctx)

    @tatsu.rule('Token')
    def token(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting('"', "'", 'r["\\\']', '→raw_string', '→string')

            @α.option
            def _(ctx: Ctx) -> Any:
                self.string(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.raw_string(ctx)

    @tatsu.rule
    def literal(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              '"',
              "'",
              '(?ms)\\s*([_\\w][_\\w\\d]*)\\b',
              '0[xX](?:\\d|[a-fA-F])+',
              'False',
              'None',
              'True',
              '[-+]?(?:\\d+\\.\\d*|\\d*\\.\\d+)(?:[Ee][-+]?\\d+)?',
              '[-+]?\\d+',
              '→boolean',
              '→false',
              '→float',
              '→hex',
              '→int',
              '→none',
              '→null',
              '→number',
              'r["\\\']',
              '→raw_string',
              '→string',
              '→true',
              '→value',
              '→word'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                self.raw_string(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.value(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.boolean(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.none(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.word(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.hex(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.float(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.int(ctx)

    @tatsu.rule
    def string(self, ctx: Ctx) -> Any:
        with ctx.if_():
            with ctx.group():
                with ctx.choice() as α:
                    α.expecting('"', "'")

                    @α.option
                    def _(ctx: Ctx) -> Any:
                        ctx.token('"')
                    @α.option
                    def _(ctx: Ctx) -> Any:
                        ctx.token("'")
        with ctx.group():
            with ctx.choice() as α:
                α.expecting('→doublequoted', '→multiline_string', '→singlequoted')

                @α.option
                def _(ctx: Ctx) -> Any:
                    self.multiline_string(ctx)
                @α.option
                def _(ctx: Ctx) -> Any:
                    self.singlequoted(ctx)
                @α.option
                def _(ctx: Ctx) -> Any:
                    self.doublequoted(ctx)

    @tatsu.rule
    def singlequoted(self, ctx: Ctx) -> Any:
        self.SINGLEQUOTED(ctx)

    @tatsu.rule
    def doublequoted(self, ctx: Ctx) -> Any:
        self.DOUBLEQUOTED(ctx)

    @tatsu.rule
    def raw_string(self, ctx: Ctx) -> Any:
        with ctx.if_():
            ctx.pattern(r'r["\']')
        ctx.pattern(r'r')
        with ctx.result():
            self.STRING(ctx)

    @tatsu.rule
    @tatsu.token
    def STRING(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              '"((?:[^"\\n]|\\\\"|\\\\\\\\)*?)"',
              "'((?:[^'\\n]|\\\\'|\\\\\\\\)*?)'",
              '→DOUBLEQUOTED',
              '→SINGLEQUOTED'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                self.SINGLEQUOTED(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.DOUBLEQUOTED(ctx)

    @tatsu.rule
    @tatsu.token
    def SINGLEQUOTED(self, ctx: Ctx) -> Any:
        ctx.pattern(r"'((?:[^'\n]|\\'|\\\\)*?)'")
        ctx.cut()

    @tatsu.rule
    @tatsu.token
    def DOUBLEQUOTED(self, ctx: Ctx) -> Any:
        ctx.pattern(r'"((?:[^"\n]|\\"|\\\\)*?)"')
        ctx.cut()

    @tatsu.rule
    def multiline_string(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              '(?ms)"""((?:\\\\\\\\|\\\\.|.)*?)"""',
              "(?ms)'''((?:\\\\\\\\|\\\\.|.)*?)'''"
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.pattern(r"(?ms)'''((?:\\\\|\\.|.)*?)'''")
                ctx.cut()
            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.pattern(r'(?ms)"""((?:\\\\|\\.|.)*?)"""')
                ctx.cut()

    @tatsu.rule
    def hex(self, ctx: Ctx) -> Any:
        ctx.pattern(r'0[xX](?:\d|[a-fA-F])+')

    @tatsu.rule
    def float(self, ctx: Ctx) -> Any:
        ctx.pattern(r'[-+]?(?:\d+\.\d*|\d*\.\d+)(?:[Ee][-+]?\d+)?')

    @tatsu.rule
    def int(self, ctx: Ctx) -> Any:
        ctx.pattern(r'[-+]?\d+')

    @tatsu.rule
    def path(self, ctx: Ctx) -> Any:
        ctx.pattern(r'[_\w][_\w\d]*(?:::[_\w][_\w\d]*)+')

    @tatsu.rule
    def word(self, ctx: Ctx) -> Any:
        ctx.pattern(r'(?ms)\s*([_\w][_\w\d]*)\b')

    @tatsu.rule('Dot')
    def dot(self, ctx: Ctx) -> Any:
        ctx.token('/./')

    @tatsu.rule('Pattern')
    def pattern(self, ctx: Ctx) -> Any:
        self.regex(ctx)

    @tatsu.rule
    def regex(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting('?/', '→deprecated_regex')

            @α.option
            def _(ctx: Ctx) -> Any:
                self.deprecated_regex(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                with ctx.ifnot_():
                    ctx.token('?/')
                with ctx.group():
                    with ctx.choice() as β:
                        β.expecting('?', '→REGEX')

                        @β.option
                        def _(ctx: Ctx) -> Any:
                            self.REGEX(ctx)
                        @β.option
                        def _(ctx: Ctx) -> Any:
                            ctx.token('?')
                            with ctx.result():
                                self.STRING(ctx)

    @tatsu.rule
    @tatsu.token
    def REGEX(self, ctx: Ctx) -> Any:
        with ctx.if_():
            ctx.token('/')
        ctx.pattern(r'(?ms)/((?:[^/\\]|\\/|\\.)*)/')
        ctx.cut()

    @tatsu.rule
    def deprecated_regex(self, ctx: Ctx) -> Any:
        ctx.token('?/')
        ctx.cut()
        with ctx.result():
            ctx.pattern(r'(?ms)((?:[^/\\]|\\/|\\.)*)')
        ctx.cut()
        ctx.token('/?')

    @tatsu.rule
    def boolean(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting('False', 'True')

            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.token('True')
            @α.option
            def _(ctx: Ctx) -> Any:
                ctx.token('False')

    @tatsu.rule
    def none(self, ctx: Ctx) -> Any:
        ctx.token('None')

    @tatsu.rule('EOF')
    def eof(self, ctx: Ctx) -> Any:
        ctx.token('$')
        ctx.cut()

    @tatsu.rule('EOL')
    def eol(self, ctx: Ctx) -> Any:
        ctx.token('$->')

    @tatsu.rule
    def value(self, ctx: Ctx) -> Any:
        with ctx.choice() as α:
            α.expecting(
              '"',
              "'",
              '-?(?:0|[1-9][0-9]*)(?:\\.[0-9]+)?(?:[eE][+-]?[0-9]+)?',
              'false',
              'null',
              '→number',
              '→string',
              'true'
            )

            @α.option
            def _(ctx: Ctx) -> Any:
                self.string(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.number(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.true(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.false(ctx)
            @α.option
            def _(ctx: Ctx) -> Any:
                self.null(ctx)

    @tatsu.rule
    def number(self, ctx: Ctx) -> Any:
        ctx.pattern(r'-?(?:0|[1-9][0-9]*)(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?')

    @tatsu.rule
    def true(self, ctx: Ctx) -> Any:
        ctx.token('true')

    @tatsu.rule
    def false(self, ctx: Ctx) -> Any:
        ctx.token('false')

    @tatsu.rule
    def null(self, ctx: Ctx) -> Any:
        ctx.token('null')


def main(filename, **kwargs):
    if not filename or filename == '-':
        import sys

        text = sys.stdin.read()
    else:
        import pathlib

        text = pathlib.Path(filename).read_text()

    parser = TatSuBootstrapParser()
    return parser.parse(text, filename=filename, **kwargs)


if __name__ == '__main__':
    import json
    from tatsu.parsing import generic_main
    from tatsu.util import asjson

    ast = generic_main(main, TatSuBootstrapParser, name='TatSuBootstrap')
    data = asjson(ast)
    print(json.dumps(data, indent=2))
